import numpy as np
import datetime
import os
import tensorflow as tf
import random
import logging
from tqdm import tqdm
import mlflow
from scipy.sparse import coo_matrix, diags

from tensorflow.keras.optimizers import Adam, SGD

from graphmb.models import  TH, TrainHelperVAE, VAEDecoder, VAEEncoder, LabelClassifier
from graphmb.utils import set_seed
from graphmb.visualize import plot_edges_sim
from graphmb.evaluate import compute_clusters_and_stats, eval_epoch
from graphmb.utils import get_cluster_mask
from graphmb.gnn_models import name_to_model
    
class TensorboardLogger(logging.StreamHandler):
    def __init__(self, file_writer, runname=""):
        logging.StreamHandler.__init__(self)
        self.file_writer = file_writer
        self.runname = runname
        self.step = 0

    def emit(self, msg):
        with self.file_writer.as_default():
            tf.summary.text(self.runname, msg.msg, step=self.step)
        self.step += 1


def log_to_tensorboard(writer, values, step):
    """Write key-values to writer
    """
    for k, v in values.items():
        with writer.as_default():
            tf.summary.scalar(k, v, step=step)


def normalize_adj_sparse(A):
    #breakpoint()
    # https://github.com/tkipf/gcn/blob/master/gcn/utils.py
    A.setdiag(1)
    rowsum = np.array(A.sum(1))
    d_inv_sqrt = np.power(rowsum, -0.5).flatten()
    d_inv_sqrt[np.isinf(d_inv_sqrt)] = 0.0
    d_mat_inv_sqrt = diags(d_inv_sqrt)
    return A.dot(d_mat_inv_sqrt).transpose().dot(d_mat_inv_sqrt).tocoo()


def prepare_data_for_gnn(
    dataset, use_edge_weights=True, cluster_markers_only=False, use_raw=False,
    binarize=False, remove_edges=False, remove_same_scg=True
):

    if use_raw: # use raw features instead of precomputed embeddings
        node_raw = np.hstack((dataset.node_depths, dataset.node_kmers))
        # features are already normalized
        #node_raw = (node_raw - node_raw.mean(axis=0, keepdims=True)) / node_raw.std(axis=0, keepdims=True)
        X = node_raw
    else:
        node_features = (dataset.node_embs - dataset.node_embs.mean(axis=0, keepdims=True)) / dataset.node_embs.std(
            axis=0, keepdims=True
        )
        X = node_features

    depth = 2
    # adjacency_matrix_sparse, edge_features = filter_graph_with_markers(adjacency_matrix_sparse, node_names, contig_genes, edge_features, depth=depth) 
    cluster_mask = get_cluster_mask(cluster_markers_only, dataset)
    #connected_marker_nodes = set(range(len(dataset.node_names)))
    
    adj_matrix = dataset.adj_matrix.copy()
    edge_weights = dataset.edge_weights.copy()
    if binarize:
        # both dataset.adj_matrix and dataset.edge_weights
        #breakpoint()
        percentile = 90
        threshold = np.percentile(dataset.edge_weights, percentile)
        print(f"using this threshold ({percentile} percentile) {threshold} on adj matrix with {len(dataset.adj_matrix.row)} edges")
        #dataset.adj_matrix = dataset.adj_matrix
        adj_matrix.data[adj_matrix.data < threshold] = 0
        adj_matrix.data[adj_matrix.data >= threshold] = 1
        adj_matrix.eliminate_zeros()
        edge_weights = adj_matrix.data 
        #dataset.edge_weights = np.ones(len(dataset.adj_matrix.row))
        print(f"reduce matrix to {len(edge_weights)} edges")
    
    if remove_same_scg:
        edges_with_same_scgs = dataset.get_edges_with_same_scgs()
        for x in edges_with_same_scgs:
                adj_matrix.data[x] = 0
                edge_weights[x] = 0
        adj_matrix.eliminate_zeros()
        print(f"deleted {len(edges_with_same_scgs)} edges with same SCGs")
    
    if remove_edges:
        # create self loops only sparse adj matrix
        n = len(dataset.node_names)
        adj_matrix = coo_matrix((np.ones(n), (np.array(range(n)), np.array(range(n)))), shape=(n,n))
        edge_weights = np.ones(len(adj_matrix.row))
        print(f"reduce matrix to {len(edge_weights)} edges")
    
    # gcn transform
    adj_norm = normalize_adj_sparse(adj_matrix)
    
    if use_edge_weights:
        #edge_features = (dataset.edge_weights - dataset.edge_weights.min()) / (
        #    dataset.edge_weights.max() - dataset.edge_weights.min()
        #)
        edge_features = edge_weights / edge_weights.max()
        # multiply normalized values by edge weights
        old_rows, old_cols = adj_matrix.row, adj_matrix.col
        old_idx_to_edge_idx = {(r, c): i for i, (r, c) in enumerate(zip(old_rows, old_cols))}
        old_values = adj_norm.data.astype(np.float32)
        new_values = []
        for i, j, ov in zip(adj_norm.row, adj_norm.col, old_values):
            if i == j:
                new_values.append(1.0)
            else:
                try:
                    eidx = old_idx_to_edge_idx[(i, j)]
                    new_values.append(ov * edge_features[eidx])
                except:
                    new_values.append(ov)
        new_values = np.array(new_values).astype(np.float32)
    
    else:
        adj_norm.data = np.ones(len(adj_norm.row))
        new_values = adj_norm.data.astype(np.float32)
    
    # convert to tf.SparseTensor
    adj = tf.SparseTensor(
        indices=np.array([adj_norm.row, adj_norm.col]).T, values=new_values, dense_shape=adj_norm.shape
    )
    adj = tf.sparse.reorder(adj)

    # neg_pair_idx = None
    pos_pair_idx = None
    print("**** Num of edges:", adj.indices.shape[0])
    return X, adj, cluster_mask, dataset.neg_pairs_idx, pos_pair_idx

def run_model_vaegnn(dataset, args, logger, nrun, plot=False, use_last_batch=False):
    set_seed(args.seed)
    node_names = np.array(dataset.node_names)
    RESULT_EVERY = args.evalepochs
    hidden_gnn = args.hidden_gnn
    hidden_vae = args.hidden_vae
    output_dim_gnn = args.embsize_gnn
    output_dim_vae = args.embsize_vae
    epochs = args.epoch
    lr_vae = args.lr_vae
    lr_gnn = args.lr_gnn
    nlayers_gnn = args.layers_gnn
    gname = args.model_name
    use_gnn = args.layers_gnn > 0
    target_metric = args.targetmetric

    with mlflow.start_run(run_name=args.assembly.split("/")[-1] + "-" + args.outname):
        mlflow.log_params(vars(args))
        
        # pick one of the LAF models
        gmodel_type = name_to_model[gname.split("_")[0].upper()] 
        clustering = args.clusteringalgo
        k = args.kclusters
        
        use_edge_weights = True
        cluster_markers_only = args.quick
        decay = 0.5 ** (2.0 / epochs)
        use_ae = True
        args.rawfeatures = True

        # setup logging
        current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        train_log_dir = os.path.join(args.outdir, 'logs/' + args.outname + current_time + '/train')
        summary_writer = tf.summary.create_file_writer(train_log_dir)
        tb_handler = TensorboardLogger(summary_writer, runname=args.outname + current_time)
        logger.addHandler(tb_handler)

        X, adj, cluster_mask, neg_pair_idx, pos_pair_idx = prepare_data_for_gnn(
            dataset, use_edge_weights, cluster_markers_only, use_raw=args.rawfeatures,
            binarize=args.binarize, remove_edges=args.noedges)

        if nrun == 0:
            print("logging to tensorboard")
            #tf.summary.trace_on(graph=True)
            logger.info("******* Running model: {} **********".format(gname))
            logger.info("***** using edge weights: {} ******".format(use_edge_weights))
            logger.info("***** cluster markers only: {} *****".format(cluster_markers_only))
            logger.info("***** threshold adj matrix: {} *****".format(args.binarize))
            logger.info("***** self edges only: {} *****".format(args.noedges))
            logger.info("***** use gnn: {} *****".format(use_gnn))
            logger.info("***** Using raw kmer+abund features: {}".format(args.rawfeatures))
            logger.info("***** SCG neg pairs: {}".format(neg_pair_idx.shape))
            logger.info("***** input features dimension: {}".format(X[cluster_mask].shape))
        tf.config.run_functions_eagerly(True)
        
        # pre train clustering
        if not args.skip_preclustering and nrun == 0:
            cluster_labels, stats, _, hq_bins = compute_clusters_and_stats(
                        X[cluster_mask], node_names[cluster_mask],
                        dataset, clustering=clustering, k=k, tsne=args.tsne,
                        amber=(args.labels is not None and "amber" in args.labels),
                        cuda=args.cuda,
                    )
            #mlflow.log_metrics(stats, step=0)
            if args.noise and hasattr(dataset, "true_adj_matrix"):
                # eval edge acc if true adj matrix exists
                # full adj matrix may be too big to compute, use only indices from true_adj
                #breakpoint()
                #row_embs = tf.gather(indices=dataset.true_adj_matrix.col, params=X)
                #col_embs = tf.gather(indices=dataset.true_adj_matrix.row, params=X)
                #positive_pairwise = tf.sigmoid(tf.reduce_sum(tf.math.multiply(row_embs, col_embs), axis=1))
                new_adj = tf.sigmoid(X @ X.transpose())
                predicted_edges = set([tuple(v) for v in tf.where(new_adj>0.5).numpy()])
                true_edges = set(zip(dataset.true_adj_matrix.col, dataset.true_adj_matrix.row))
                correct_edges = predicted_edges & true_edges
                print(len(correct_edges), len(true_edges), len(predicted_edges))
            logger.info(f">>> Pre train stats: {str(stats)}")
        else:
            stats = {"epoch":0, target_metric: 0, "hq": 0 }
            cluster_labels = []
        
        pname = ""

        #plot edges vs initial embs
        if plot:
            id_to_scg = {i: set(dataset.contig_markers[node_name].keys()) for i, node_name in enumerate(dataset.node_names)}
            plot_edges_sim(X, dataset.adj_matrix, id_to_scg, f"{args.outdir}/{args.outname}_pretrain_")

        # initialize variables
        scores = [stats]
        all_cluster_labels = []
        all_cluster_labels.append(cluster_labels)
        losses = {"total": [], "ae": [], "gnn": [], "scg": []}
        X = X.astype(np.float32)
        features = tf.constant(X)
        input_dim_gnn = output_dim_vae #+ dataset.node_depths.shape[1]
        if args.gnn_alpha > 0:
            clustering_dim = output_dim_gnn
        elif not args.concatfeatures:
            clustering_dim = output_dim_vae
        else:
            clustering_dim = output_dim_vae + output_dim_gnn
        logger.info(f"*** Model input dim {X.shape[1]}, GNN input dim {input_dim_gnn}")
        logger.info(f"use_ae: {use_ae}, run AE only: {args.ae_only} output clustering dim {clustering_dim}")

        gnn_model = gmodel_type(
            features_shape=features.shape,
            input_dim=input_dim_gnn,
            labels=None,
            adj=adj,
            embsize=output_dim_gnn,
            hidden_units=hidden_gnn,
            layers=nlayers_gnn,
            conv_last=False,
            dropout=args.dropout_gnn
        )  # , use_bn=True, use_vae=False)

        encoder = VAEEncoder(dataset.node_depths.shape[1], dataset.node_kmers.shape[1],
                             hidden_vae, zdim=output_dim_vae, dropout=args.dropout_vae,
                             layers=args.layers_vae)
        decoder = VAEDecoder(dataset.node_depths.shape[1], dataset.node_kmers.shape[1],
                             hidden_vae, zdim=output_dim_vae, dropout=args.dropout_vae,
                             layers=args.layers_vae)

        if args.classify:
            classifier = LabelClassifier(len(dataset.labels), zdim=encoder.zdim)
        else:
            classifier = None
        trainer = TH(
            features,
            gnn_model=gnn_model,
            lr=lr_gnn,
            all_different_idx=neg_pair_idx,
            all_same_idx=pos_pair_idx,
            ae_encoder=encoder,
            ae_decoder=decoder,
            decoder_input=args.decoder_input,
            classifier=classifier,
            latentdim=output_dim_gnn,
            gnn_weight=float(args.gnn_alpha),
            ae_weight=float(args.ae_alpha),
            scg_weight=float(args.scg_alpha),
            num_negatives=args.negatives,
            kmers_dim=dataset.node_kmers.shape[1],
            abundance_dim=dataset.node_depths.shape[1],
            use_gnn=use_gnn,
            use_noise=args.noise,
            loglevel=args.loglevel,
            pretrainvae=args.vaepretrain
        )

        if args.batchtype == "auto":
            if args.ae_alpha > 0 and args.gnn_alpha == 0 and args.scg_alpha == 0:
                args.batchtype = "nodes"
            else:
                args.batchtype = "edges"
        scg_idx = np.arange(len(neg_pair_idx))
        # create eval split
        if args.eval_split == 0:
            if "node" in args.batchtype:
                train_idx = np.arange(len(features))
            elif "edge" in args.batchtype:
                train_idx = np.arange(gnn_model.adj.indices.shape[0])
            eval_idx = []
        else:
            if "node" in args.batchtype:
                train_idx = np.array(random.sample(list(range(gnn_model.adj.indices.shape[0])), int(gnn_model.adj.indices.shape[0]*(1-args.eval_split))))
                eval_idx = np.array([x for x in np.arange(gnn_model.adj.indices.shape[0]) if x not in train_idx])
            elif "edge" in args.batchtype:
                train_idx = np.array(random.sample(list(range(gnn_model.adj.indices.shape[0])), int(gnn_model.adj.indices.shape[0]*(1-args.eval_split))))
                eval_idx = np.array([x for x in np.arange(gnn_model.adj.indices.shape[0]) if x not in train_idx])
            logging.info(f"**** using {gnn_model.adj.indices.shape[0]} for training and {gnn_model.adj.indices.shape[0]} for eval")
        #edges_idx = np.arange(gnn_model.adj.indices.shape[0])
        features = np.array(features)
        pbar_epoch = tqdm(range(epochs), disable=args.quiet, position=0)
        scores = [stats]
        best_embs, best_vae_embs, best_model, best_score, best_epoch = None, None, None, 0, 0
        
        # increasing batch size
        batch_size = args.batchsize
        if batch_size == 0:
            batch_size = len(train_idx)
        logger.info("**** initial {} batch size: {} ****".format(args.batchtype, batch_size))
        mlflow.log_metric("cur_batch_size", batch_size, 0)
        batch_steps = [25, 75, 150, 300]
        batch_steps = [x for i, x in enumerate(batch_steps) if (2 ** (i+1))*batch_size < len(train_idx)]
        logger.info("**** epoch batch size doubles: {} ****".format(str(batch_steps)))
        gold_labels = np.array([dataset.labels.index(dataset.node_to_label[n]) for n in dataset.node_names])
        step = 0
        for e in pbar_epoch:
            #vae_epoch_losses = {"kld_loss": [], "vae_loss": [], "kmer_loss": [], "ab_loss": []}
            recon_loss = 0
            trainer.epoch = e
            #if args.vaepretrain - 1 == e:
            #    trainer.opt = Adam(learning_rate=lr_gnn/10, epsilon=1e-8)
            # train VAE in batches
            if e in batch_steps:
                print(f'Increasing {args.batchtype} batch size from {batch_size:d} to {batch_size*2:d}')
                batch_size = batch_size * 2
                mlflow.log_metric("cur_batch_size", batch_size, e )

            ############################################################
            vae_losses_epoch = {}
            gnn_losses_epoch = {}
            total_losses_epoch = []

            # train model ################################################
            np.random.shuffle(train_idx)

            #graph_batch_size = 256
            #graph_batch_size = len(edges_idx)
            n_batches = len(train_idx)//batch_size
            if use_last_batch and n_batches < len(train_idx)/batch_size:
                n_batches += 1 # add final batch
            if args.scg_alpha > 0:
                scg_batch_size = len(neg_pair_idx)//n_batches
                logging.info("*** scg batch size: {}".format(scg_batch_size))
            pbar_gnnbatch = tqdm(range(n_batches), disable=(args.quiet or batch_size == len(train_idx) or n_batches < 1000), position=1, ascii=' =')
            for b in pbar_gnnbatch:
                if "edge" in args.batchtype:
                    nodes_batch, edges_batch = None, train_idx[b*batch_size:(b+1)*batch_size]
                elif "node" in args.batchtype:
                    nodes_batch, edges_batch = train_idx[b*batch_size:(b+1)*batch_size], None
                if args.scg_alpha > 0:
                    scg_batch = scg_idx #[b*scg_batch_size:(b+1)*scg_batch_size]
                else:
                    scg_batch = None
                losses = trainer.train_unsupervised(edges_idx=edges_batch,
                                                    nodes_idx=nodes_batch,
                                                    scgs_idx=scg_batch, vae=True,
                                                    gold_labels=gold_labels)
                total_loss, gnn_losses, ae_losses = losses
                total_losses_epoch.append(total_loss.numpy())
                for l in gnn_losses: gnn_losses_epoch.setdefault(l,[]).append(gnn_losses[l])
                for l in ae_losses: vae_losses_epoch.setdefault(l,[]).append(ae_losses[l])

                #add losses to get epoch avg
            if args.scg_alpha > 0:
                scg_loss = trainer.train_scg_only()
                scg_loss = scg_loss.numpy()
            else:
                scg_loss = 0
            epoch_metrics = {"Total": np.average(total_losses_epoch),
                             "gnn": np.average(gnn_losses_epoch["gnn_loss"]),
                             #"SCG": np.average(gnn_losses_epoch["scg_loss"]),
                             "SCG": scg_loss,
                             "pred": np.average(gnn_losses_epoch["pred_loss"]),
                            #'GNN  LR': float(trainer.opt._decayed_lr(float)),
                            "pos": np.average(gnn_losses_epoch["pos"]),
                            "neg": np.average(gnn_losses_epoch["neg"]),
                            "kld": np.average(vae_losses_epoch["kld"]),
                            "vae": np.average(vae_losses_epoch["vae_loss"]),
                            "kmer": np.average(vae_losses_epoch["kmer_loss"]),
                            "ab": np.average( vae_losses_epoch["ab_loss"])}
                            #"logvar": ae_losses["mean_logvar"],
                            #"grad_norm": ae_losses["grad_norm"],
                            #"grad_norm_clip": ae_losses["grad_norm_clip"]}
            #print(float(gnn_trainer.opt._decayed_lr(float)))
            log_to_tensorboard(summary_writer, epoch_metrics, step)
            mlflow.log_metrics(epoch_metrics, step=e)
            ##############################################################


            # eval loss:TODO #######################################################

            gpu_mem_alloc = tf.config.experimental.get_memory_info('GPU:0')["peak"] / 1000000 if args.cuda else 0
            #gpu_mem_alloc = tf.config.experimental.get_memory_usage('GPU:0') / 1000000 if args.cuda else 0
            # eval checkpoint ##############################################################
            if (e + 1) % RESULT_EVERY == 0 and e > args.evalskip:
                evalstarttime = datetime.datetime.now()
               
                gnn_input_features = trainer.encoder(features)[0]
                #trainer.features = gnn_input_features
                logger.debug("encoder output " + str(gnn_input_features[0][:5].numpy()))
                if use_gnn: # and e > args.vaepretrain:
                    node_new_features = trainer.gnn_model(gnn_input_features, None, training=False)
                    
                    if args.concatfeatures:
                        node_new_features = tf.concat((gnn_input_features,
                                                        node_new_features), axis=1).numpy()
                    else:
                        node_new_features = node_new_features.numpy()
                        #node_new_features = gnn_input_features.numpy()
                else:
                    node_new_features = gnn_input_features.numpy()
                if args.noise:
                    #mlflow.log_metrics({"avg positive noise": trainer.positive_noises.numpy().mean(),
                    #                               "avg scg noise": trainer.scg_noises.numpy().mean()}, step=e)
                    # get topk edge noises
                    #breakpoint()
                    #topk_indices = tf.math.top_k(tf.math.abs(trainer.positive_noises[:, 0]), k=10).indices
                    sorted_noises = sorted(trainer.noise_dict.items(), key=lambda item: item[1])
                    topk_indices = sorted_noises[-5:]
                    bottomk_indices = sorted_noises[:5]
                    #breakpoint()
                    logger.debug("            src (label) dst (label) observed predicted noise")
                    for pair, noise in topk_indices + bottomk_indices:
                        logger.debug("{} ({}) {} ({}) 0 {:.4f} {}".format(dataset.node_names[pair[0]],
                                     dataset.node_to_label[dataset.node_names[pair[0]]],
                                     dataset.node_names[pair[1]], 
                                     dataset.node_to_label[dataset.node_names[pair[1]]],
                                     #gnn_model.adj.values[i].numpy(),
                                     tf.reduce_sum(tf.math.multiply(tf.nn.l2_normalize(node_new_features[pair[0]]),
                                                                    tf.nn.l2_normalize(node_new_features[pair[1]]))).numpy(),
                                    noise))


                if args.noise and hasattr(dataset, "true_adj_matrix"):
                    # eval edge acc if true adj matrix exists
                    # full adj matrix may be too big to compute, use only indices from true_adj
                    breakpoint()
                    #row_embs = tf.gather(indices=dataset.true_adj_matrix.col, params=node_new_features)
                    #col_embs = tf.gather(indices=dataset.true_adj_matrix.row, params=node_new_features)
                    #positive_pairwise = tf.sigmoid(tf.reduce_sum(tf.math.multiply(row_embs, col_embs), axis=1))
                    new_adj = tf.sigmoid(node_new_features @ node_new_features.transpose())
                    predicted_edges = set([tuple(v) for v in tf.where(new_adj>0.5).numpy()])
                    true_edges = set(zip(dataset.true_adj_matrix.col, dataset.true_adj_matrix.row))
                    correct_edges = predicted_edges & true_edges
                    print(len(correct_edges), len(true_edges), len(predicted_edges))
                    #correct_edges = dataset.true_adj_matrix.indices & 
                    #pass
                    
                weights = (trainer.encoder.get_weights(), trainer.gnn_model.get_weights())
                best_score, best_embs, best_epoch, scores, best_model, cluster_labels = eval_epoch(logger, summary_writer,
                                                                    node_new_features, cluster_mask, weights,
                                                                    step, args, dataset, e, scores,
                                                                    best_score, best_embs, best_epoch, best_model, target_metric)
                stats = scores[-1]
                all_cluster_labels.append(cluster_labels)
                if args.quiet:
                    logger.info(f"--- EPOCH {e:d} ---")
                    scores_string = f"HQ={stats['hq']}  Best{target_metric}={round(best_score, 3)} Best Epoch={best_epoch} Cur={round(stats.get(target_metric,0), 3)}"
                    losses_string = " ".join([f"{k}={v:.3f}" for k, v in epoch_metrics.items()])
                    logger.info(f"[{args.outname} {nlayers_gnn}l {pname}]{losses_string} {scores_string} GPU={gpu_mem_alloc:.1f}MB")
                    logger.info(str(stats))
                mlflow.log_metrics(stats, step=e)
                #print("total eval time", datetime.datetime.now() - evalstarttime)
            losses_string = " ".join([f"{k}={v:.3f}" for k, v in epoch_metrics.items() if v != 0])
            scores_string = f"Best{target_metric}={round(best_score, 3)} BestEpoch={best_epoch} Cur{target_metric}={round(stats.get(target_metric,0), 3)}"
            pbar_epoch.set_description(
                f"[{args.outname} {pname}] {losses_string} {scores_string} GPU={gpu_mem_alloc:.1f}MB"
            )

        #################################################################
        if best_embs is None:
            best_embs = node_new_features
        
        cluster_labels, stats, _, _ = compute_clusters_and_stats(
            best_embs, node_names, dataset, clustering=clustering, k=k,
            amber=(args.labels is not None and "amber" in args.labels),
            cuda=args.cuda,
        )
        all_cluster_labels.append(cluster_labels)
        stats["epoch"] = e
        scores.append(stats)
        # get best stats:
        target_scores = [s[target_metric] for s in scores]
        best_idx = np.argmax(target_scores)
        mlflow.log_metrics(scores[best_idx], step=e+1)

        logger.info(f">>> Last epoch: {RESULT_EVERY + (best_idx*RESULT_EVERY)} : {stats} <<<")
        logger.info(f">>> best epoch: {RESULT_EVERY + (best_idx*RESULT_EVERY)} : {scores[best_idx]} <<<")
        with open(f"{args.outdir}/{args.outname}_{nrun}_best_contig2bin.tsv", "w") as f:
            f.write("@Version:0.9.0\n@SampleID:SAMPLEID\n@@SEQUENCEID\tBINID\n")
            for i in range(len(all_cluster_labels[best_idx])):
                f.write(f"{node_names[i]}\t{all_cluster_labels[best_idx][i]}\n")
     
        #plot edges vs initial embs
        #plot_edges_sim(best_vae_embs, dataset.adj_matrix, id_to_scg, "vae_")
        if plot:
            plot_edges_sim(best_embs, dataset.adj_matrix, id_to_scg, "posttrain_")
        return best_embs, scores[best_idx]


import time
import pdb
import os
import random
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import dgl
from dgl.nn.pytorch.conv import SAGEConv, GATConv
import dgl.function as fn
import tqdm
import numpy as np
import copy
import sklearn

from graphmb.utils import set_seed
from graphmb.graphmb1 import cluster_eval

# Based on this implemention: https://github.com/dmlc/dgl/blob/master/examples/pytorch/graphsage/train_sampling_unsupervised.py


class MultiLayerNeighborWeightedSampler(dgl.dataloading.MultiLayerNeighborSampler):
    def sample_frontier(self, block_id, g, seed_nodes):
        fanout = self.fanouts[block_id]
        frontier = dgl.sampling.sample_neighbors(g, seed_nodes, fanout, replace=self.replace, prob="weight")
        return frontier


class NegativeSampler(object):
    def __init__(self, g, k, neg_share=False):
        self.weights = g.in_degrees().float() ** 0.75
        self.k = k
        self.neg_share = neg_share

    def __call__(self, g, eids):
        src, _ = g.find_edges(eids)
        n = len(src)
        if self.neg_share and n % self.k == 0:
            dst = self.weights.multinomial(n, replacement=True)
            dst = dst.view(-1, 1, self.k).expand(-1, self.k, -1).flatten()
        else:
            dst = self.weights.multinomial(n * self.k, replacement=True)
        src = src.repeat_interleave(self.k)
        return src, dst


class NegativeSamplerWeight(object):
    """Samples negatives according to the inverse of edge weight"""

    def __init__(self, g, k, neg_share=False):
        self.k = k
        self.neg_share = neg_share

    def __call__(self, g, eids):
        src, dst = g.find_edges(eids)
        new_dst = []
        new_src = []
        possible_dst = set(dst)
        new_src = [src_node for src_node in src if len(g.out_edges(src_node)[1]) < 2]
        new_dst = random.sample(possible_dst, k=len(new_src))
        # print("adding", len(new_src), "more edges")
        src = src.repeat_interleave(self.k)
        expand_g = dgl.remove_self_loop(dgl.add_edges(g, new_src, new_dst))
        expand_g.edata["weight"] = 1 / expand_g.edata["weight"]

        samples_edges = dgl.sampling.sample_neighbors(
            expand_g, src, self.k, prob="weight", replace=True, edge_dir="out"
        )
        src = samples_edges.edges()[0]
        dst = samples_edges.edges()[1]
        # src = src.repeat_interleave(self.k)
        return src, dst


class CrossEntropyLoss(nn.Module):
    def forward(self, block_outputs, pos_graph, neg_graph, weights=True):
        with pos_graph.local_scope():
            pos_graph.ndata["h"] = block_outputs
            pos_graph.apply_edges(fn.u_dot_v("h", "h", "score"))
            pos_score = pos_graph.edata["score"]
        with neg_graph.local_scope():
            neg_graph.ndata["h"] = block_outputs
            neg_graph.apply_edges(fn.u_dot_v("h", "h", "score"))
            neg_score = neg_graph.edata["score"]

        score = torch.cat([pos_score, neg_score])
        pos_label = torch.ones_like(pos_score)
        if weights:
            pos_weights = pos_graph.edata["weight"] / max(pos_graph.edata["weight"])
            pos_weights = pos_weights.unsqueeze(-1)
        else:
            pos_weights = torch.ones_like(pos_score)
        neg_label = torch.zeros_like(neg_score)
        neg_weights = torch.ones_like(neg_score)
        all_weights = torch.cat([pos_weights, neg_weights])
        label = torch.cat([pos_label, neg_label]).long()
        loss = F.binary_cross_entropy_with_logits(score, label.float(), weight=all_weights)
        return loss


class SAGE(nn.Module):
    def __init__(self, in_feats, n_hidden, n_classes, n_layers, activation, dropout, agg="mean"):
        super().__init__()
        self.init(in_feats, n_hidden, n_classes, n_layers, activation, dropout, agg)

    def init(self, in_feats, n_hidden, n_classes, n_layers, activation, dropout, agg):
        self.n_layers = n_layers
        self.n_hidden = n_hidden
        self.n_classes = n_classes
        self.layers = nn.ModuleList()
        # self.batchnorm = nn.BatchNorm1d(n_hidden)
        if n_layers > 1:
            self.layers.append(SAGEConv(in_feats, n_hidden, agg, feat_drop=dropout))

            for i in range(1, n_layers - 1):
                self.layers.append(SAGEConv(n_hidden, n_hidden, agg, feat_drop=dropout))
                # self.layers.append(nn.BatchNorm1d(n_hidden))
            self.layers.append(SAGEConv(n_hidden, n_classes, agg))
        else:
            self.layers.append(SAGEConv(in_feats, n_classes, agg, feat_drop=dropout))
        if agg == "gcn":
            breakpoint()
        self.dropout = nn.Dropout(dropout)
        self.activation = activation

    def forward(self, blocks, x, edge_weight=None):
        h = x
        for il, (layer, block) in enumerate(zip(self.layers, blocks)):
            # weights = F.softmax(block.edata["weight"])
            # weights = block.edata["weight"] / max(block.edata["weight"])
            # h = layer(block, h, edge_weight=weights)
            h = layer(block, h)
            if il != len(self.layers) - 1:
                h = self.activation(h)
        return h

    def inference(self, g, x, device, batch_size, num_workers, use_weights=False):
        """
        Inference with the GraphSAGE model on full neighbors (i.e. without neighbor sampling).
        g : the entire graph.
        x : the input of entire node set.
        The inference code is written in a fashion that it could handle any number of nodes and
        layers.
        """
        # During inference with sampling, multi-layer blocks are very inefficient because
        # lots of computations in the first few layers are repeated.
        # Therefore, we compute the representation of all nodes layer by layer.  The nodes
        # on each layer are of course splitted in batches.
        # TODO: can we standardize this?
        for il, layer in enumerate(self.layers):
            y = torch.zeros(g.num_nodes(), self.n_hidden if il != len(self.layers) - 1 else self.n_classes)
            sampler = dgl.dataloading.MultiLayerFullNeighborSampler(1)
            
            dataloader = dgl.dataloading.NodeDataLoader(
                g,
                torch.arange(g.num_nodes()).to(g.device),
                sampler,
                batch_size=batch_size,
                shuffle=False,
                drop_last=False,
                num_workers=num_workers,
            )

            for input_nodes, output_nodes, blocks in dataloader:
                block = blocks[0]

                block = block.int().to(device)
                h = x[input_nodes].to(device)

                if use_weights:
                    weights = block.edata["weight"] / max(block.edata["weight"])
                    h = layer(block, h, edge_weight=weights)
                else:
                    h = layer(block, h)
                if il != len(self.layers) - 1:
                    h = self.activation(h)
                    h = self.dropout(h)

                y[output_nodes] = h.cpu()

            x = y
        return y


def train_graphsage(
    dataset,
    model,
    batch_size,
    fan_out,
    num_negs,
    neg_share,
    lr,
    num_epochs,
    num_workers=0,
    print_interval=3,
    device="cpu",
    cluster_features=True,
    clusteringalgo="kmeans",
    k=1,
    logger=None,
    loss_weights=False,
    sample_weights=False,
    epsilon=0.1,
    evalepochs=1,
    seed=0,
    eval_skip=0,
    quick=False
):
    train_start_time = time.time()
    nfeat = dataset.graph.ndata.pop("feat")
    model = model.to(device)
    # Create PyTorch DataLoader for constructing blocks
    n_edges = dataset.graph.num_edges()
    train_seeds = torch.arange(n_edges)
    set_seed(seed)

    # Create samplers
    if not sample_weights:
        neg_sampler = NegativeSampler(dataset.graph, num_negs, neg_share)
        sampler = dgl.dataloading.MultiLayerNeighborSampler([int(fanout) for fanout in fan_out.split(",")])
    else:
        neg_sampler = NegativeSamplerWeight(dataset.graph, num_negs, neg_share)
        sampler = MultiLayerNeighborWeightedSampler([int(fanout) for fanout in fan_out.split(",")])

    if batch_size == 0:
        batch_size = len(train_seeds)
    dataloader = dgl.dataloading.EdgeDataLoader(
        dataset.graph,
        train_seeds,
        sampler,
        exclude="reverse_id",
        reverse_eids=torch.cat([torch.arange(n_edges // 2, n_edges), torch.arange(0, n_edges // 2)]).to(train_seeds),
        negative_sampler=neg_sampler,
        device=device,
        batch_size=batch_size,
        shuffle=True,
        drop_last=True,
        num_workers=num_workers,
    )

    loss_fcn = CrossEntropyLoss()
    optimizer_sage = optim.Adam(model.parameters(), lr=lr)
    # Training loop
    avg = 0
    iter_pos = []
    iter_neg = []
    iter_d = []
    iter_t = []
    best_hq = 0
    best_hq_epoch = 0
    total_steps = 0
    losses = []
    stop_run = False
    for epoch in range(num_epochs):
        tic = time.time()
        # Loop over the dataloader to sample the computation dependency graph as a list of
        # blocks.
        tic_step = time.time()
        for step, (input_nodes, pos_graph, neg_graph, blocks) in enumerate(dataloader):
            batch_inputs = nfeat[input_nodes].to(device)
            d_step = time.time()
            #set_seed(seed)
            model.train()
            pos_graph = pos_graph.to(device)
            neg_graph = neg_graph.to(device)
            blocks = [block.int().to(device) for block in blocks]
            # Compute loss and prediction
            batch_pred = model(blocks, batch_inputs)
            loss = loss_fcn(batch_pred, pos_graph, neg_graph, weights=loss_weights)

            optimizer_sage.zero_grad()
            loss.backward()
            optimizer_sage.step()

            t = time.time()
            pos_edges = pos_graph.num_edges()
            neg_edges = neg_graph.num_edges()
            iter_pos.append(pos_edges / (t - tic_step))
            iter_neg.append(neg_edges / (t - tic_step))
            iter_d.append(d_step - tic_step)
            iter_t.append(t - d_step)
            if total_steps % print_interval == 0:
                gpu_mem_alloc = torch.cuda.max_memory_allocated() / 1000000 if torch.cuda.is_available() else 0
                logger.info(
                    "Epoch {:05d} | Step {:05d} | N samples {} | Loss {:.4f} | Speed (samples/sec) {:.4f}|{:.4f} | Load {:.4f}| train {:.4f} | GPU {:.1f} MB".format(
                        epoch,
                        step,
                        len(input_nodes),
                        loss.item(),
                        np.mean(iter_pos[3:]),
                        np.mean(iter_neg[3:]),
                        np.mean(iter_d[3:]),
                        np.mean(iter_t[3:]),
                        gpu_mem_alloc,
                    )
                )
            tic_step = time.time()
            total_steps += 1

        losses.append(loss.item())
        # early stopping
        if (
            epsilon is not None
            and len(losses) > 3
            and (losses[-2] - losses[-1]) < epsilon
            and (losses[-3] - losses[-2]) < epsilon
        ):
            logger.info("Early stopping {}".format(str(losses[-5:])))
            stop_run = True

        if dataset.assembly.ref_marker_sets is not None and ((epoch % evalepochs == 0 and epoch > eval_skip) or stop_run):
            model.eval()
            encoded = model.inference(dataset.graph, nfeat, device, batch_size, num_workers)
            if cluster_features:
                encoded = torch.cat((encoded, nfeat), axis=1)
            best_hq, best_hq_epoch, kmeans_loss, clusters, epoch_metrics = cluster_eval(
                model=model,
                dataset=dataset.assembly,
                logits=encoded,
                clustering=clusteringalgo,
                k=k,
                loss=loss,
                best_hq=best_hq,
                best_hq_epoch=best_hq_epoch,
                epoch=epoch,
                device=device,
                clusteringloss=False,
                logger=logger,
                use_marker_contigs_only=quick,
                seed=seed,
            )
            if best_hq_epoch == epoch:
                metrics = epoch_metrics
            # compare clusters
            new_assignments = np.zeros(len(dataset.assembly.node_names))
            for i, cluster in enumerate(clusters):
                for contig in clusters[cluster]:
                    new_assignments[dataset.assembly.node_names.index(contig)] = i

            old_assignments = new_assignments.copy()
        toc = time.time()
        if epoch >= 5:
            avg += toc - tic
        if stop_run:
            break
        # = encoded.cpu().detach().numpy()

    last_train_embs = encoded
    last_model = model
    logger.info("saving last model")
    torch.save(last_model.state_dict(), os.path.join(dataset.assembly.cache_dir, "last_model_hq.pkl"))
    logger.info("Avg epoch time: {}".format(avg / (epoch - 4)))
    logger.info("Total training time: {:.3f} seconds".format(time.time() - train_start_time))
    logger.info("Peak memory usage: {} MB".format(torch.cuda.max_memory_allocated()/1000000))
    model.eval()
    logger.info(f"Best HQ {best_hq} epoch, {best_hq_epoch}")
    if total_steps == 0:
        print("No training was done")
    elif dataset.assembly.ref_marker_sets is not None:
        logger.info("loading best model")
        best_model = copy.deepcopy(model)
        try:
            best_model.load_state_dict(torch.load(os.path.join(dataset.assembly.cache_dir, "best_model_hq.pkl")))
        except RuntimeError:
            pdb.set_trace()
    else:
        best_model = last_model
    set_seed(seed)
    print("running best or last model again")
    best_train_embs = best_model.inference(dataset.graph, nfeat, device, batch_size, num_workers)

    best_train_embs = best_train_embs.detach()
    if cluster_features:
        best_train_embs = torch.cat((best_train_embs, nfeat), axis=1).detach()
    return best_train_embs, best_model, last_train_embs, last_model, metrics



import argparse

class LoadFromFile (argparse.Action):
    def __call__ (self, parser, namespace, values, option_string = None):
        with values as f:
            contents = f.read()
        # parse arguments in the file and store them in a blank namespace
        data = parser.parse_args(contents.split(), namespace=None)
        for k, v in vars(data).items():
            # set arguments in the target namespace if they haven’t been set yet
            #if getattr(namespace, k, None) is None:
            #print(f"using args {k}={v}")
            setattr(namespace, k, v)

def create_parser():
    parser = argparse.ArgumentParser(description="Train graph embedding model")
    # input files
    parser.add_argument("--assembly", type=str, help="Assembly base path", required=False)
    parser.add_argument("--assembly_name", type=str, help="File name with contigs", default="assembly.fasta")
    parser.add_argument("--graph_file", type=str, help="File name with graph", default="assembly_graph.gfa")
    parser.add_argument(
        "--edge_threshold", type=float, help="Remove edges with weight lower than this (keep only >=)", default=None
    )
    parser.add_argument("--depth", type=str, help="Depth file from jgi", default="assembly_depth.txt")
    parser.add_argument(
        "--features", type=str, help="Features file mapping contig name to features", default="features.tsv"
    )
    parser.add_argument("--labels", type=str, help="File mapping contig to label", default=None)
    parser.add_argument("--embs", type=str, help="No train, load embs", default=None)

    # model specification
    parser.add_argument("--model_name", type=str, help="One of the implemented models", default="gcn")
    parser.add_argument(
        "--activation", type=str, help="Activation function to use(relu, prelu, sigmoid, tanh)", default="relu"
    )
    parser.add_argument("--layers_vae", type=int, help="Number of layers of the VAE", default=2)
    parser.add_argument("--layers_gnn", type=int, help="Number of layers of the GNN", default=3)
    parser.add_argument("--hidden_gnn", type=int, help="Dimension of hidden layers of GNN", default=128)
    parser.add_argument("--hidden_vae", type=int, help="Dimension of hidden layers of VAE", default=512)
    parser.add_argument("--embsize_gnn", "--zg", type=int, help="Output embedding dimension of GNN", default=32)
    parser.add_argument("--embsize_vae", "--zl", type=int, help="Output embedding dimension of VAE", default=32)
    parser.add_argument("--batchsize", type=int, help="batchsize to train the VAE", default=256)
    parser.add_argument("--batchtype", type=str, help="Batch type, nodes or edges", default="auto")
    parser.add_argument("--dropout_gnn", type=float, help="dropout of the GNN", default=0.1)
    parser.add_argument("--dropout_vae", type=float, help="dropout of the VAE", default=0.2)
    parser.add_argument("--lr_gnn", type=float, help="learning rate", default=1e-2)
    parser.add_argument("--lr_vae", type=float, help="learning rate", default=1e-3)
    parser.add_argument("--gnn_alpha", type=float, help="Coeficient for GNN loss", default=0.1)
    parser.add_argument("--kld_alpha", type=float, help="Coeficient for KLD loss", default=200)
    parser.add_argument("--ae_alpha", type=float, help="Coeficient for AE loss", default=1)
    parser.add_argument("--scg_alpha", type=float, help="Coeficient for SCG loss", default=0.1)
    parser.add_argument("--clusteringalgo", help="clustering algorithm", default="vamb")
    parser.add_argument("--kclusters", help="Number of clusters (only for some clustering methods)", default=None)
    # GraphSAGE params
    parser.add_argument("--aggtype", help="Aggregation type for GraphSAGE (mean, pool, lstm, gcn)", default="lstm")
    parser.add_argument("--decoder_input", help="What to use for input to the decoder", default="vae")
    parser.add_argument("--vaepretrain", help="How many epochs to pretrain VAE", default=0, type=int)
    parser.add_argument("--ae_only", help="Do not use GNN (ae model must be used and decoder input must be ae", action="store_true")
    parser.add_argument("--negatives", help="Number of negatives to train GraphSAGE", default=10, type=int)
    parser.add_argument("--quick", help="Reduce number of nodes to run quicker", action="store_true")
    parser.add_argument("--classify", help="Run classification instead of clustering", action="store_true")
    parser.add_argument(
        "--fanout", help="Fan out, number of positive neighbors sampled at each level", default="10,25"
    )
    # other training params
    parser.add_argument("--epoch", type=int, help="Number of epochs to train model", default=200)
    parser.add_argument("--print", type=int, help="Print interval during training", default=10)
    parser.add_argument("--evalepochs", type=int, help="Epoch interval to run eval", default=20)
    parser.add_argument("--evalskip", type=int, help="Skip eval of these epochs", default=50)
    parser.add_argument("--eval_split", type=float, help="Percentage of dataset to use for eval", default=0.0)
    parser.add_argument("--kmer", default=4)
    parser.add_argument("--rawfeatures", help="Use raw features", action="store_true")
    parser.add_argument("--clusteringloss", help="Train with clustering loss", action="store_true")
    parser.add_argument("--targetmetric", help="Metric to pick best epoch", default="hq")
    parser.add_argument(
        "--concatfeatures", help="Concat learned and original features before clustering", action="store_true"
    )
    parser.add_argument("--no_loss_weights", action="store_false", help="Using edge weights for loss (positive only)")
    parser.add_argument("--no_sample_weights", action="store_false", help="Using edge weights to sample negatives")
    parser.add_argument(
        "--early_stopping",
        type=float,
        help="Stop training if delta between last two losses is less than this",
        default="0.1",
    )
    parser.add_argument("--nruns", type=int, help="Number of runs", default=1)
    # data processing
    parser.add_argument("--mincontig", type=int, help="Minimum size of input contigs", default=1000)
    parser.add_argument("--minbin", type=int, help="Minimum size of clusters in bp", default=200000)
    parser.add_argument("--mincomp", type=int, help="Minimum size of connected components", default=1)
    parser.add_argument("--randomize", help="Randomize graph", action="store_true")
    parser.add_argument("--labelgraph", help="Create graph based on labels (ignore assembly graph)", action="store_true")
    parser.add_argument("--binarize", help="Binarize adj matrix", action="store_true")
    parser.add_argument("--noedges", help="Remove all but self edges from adj matrix", action="store_true")
    parser.add_argument("--read_embs", help="Read embeddings from file", action="store_true")
    parser.add_argument("--reload", help="Reload data", action="store_true")

    parser.add_argument("--markers", type=str, help="File with precomputed checkm results to eval",
                        default="marker_gene_stats.tsv")
    parser.add_argument("--post", help="Output options", default="writeembs_contig2bin")
    parser.add_argument("--skip_preclustering", help="Use precomputed checkm results to eval", action="store_true")
    parser.add_argument("--outname", "--outputname", help="Output (experiment) name", default="")
    parser.add_argument("--cuda", help="Use gpu", action="store_true")
    parser.add_argument("--noise", help="Use noise generator", action="store_true")
    parser.add_argument("--vamb", help="Run vamb instead of loading features file", action="store_true")
    parser.add_argument("--savemodel", help="Save best model to disk", action="store_true")
    parser.add_argument("--tsne", help="Plot tsne at checkpoints", action="store_true")
    parser.add_argument("--vambdim", help="VAE latent dim", default=32)
    parser.add_argument("--numcores", help="Number of cores to use", default=1, type=int)
    parser.add_argument(
        "--outdir", "--outputdir", help="Output dir (same as input assembly dir if not defined", default=None
    )
    parser.add_argument("--assembly_type", help="flye or spades", default="flye")
    parser.add_argument("--contignodes", help="Use contigs as nodes instead of edges", action="store_true")
    parser.add_argument("--seed", help="Set seed", default=1, type=int)
    parser.add_argument("--quiet", "-q", help="Do not output epoch progress", action="store_true")
    parser.add_argument("--read_cache", help="Do not check assembly files, read cached files only", action="store_true")
    parser.add_argument("--version", "-v", help="Print version and exit", action="store_true")
    parser.add_argument("--loglevel", "-l", help="Log level", default="info")
    parser.add_argument('--configfile', type=open, action=LoadFromFile)
    return parser



import numpy as np
import datetime
import os
import tensorflow as tf
import random
import logging
from tqdm import tqdm
import mlflow
from graphmb.models import  TH
from graphmb.utils import set_seed # , run_tsne, plot_embs, plot_edges_sim
from graphmb.evaluate import calculate_overall_prf
from graphmb.train_vaegnn import prepare_data_for_gnn, TensorboardLogger, log_to_tensorboard
from graphmb.visualize import plot_edges_sim
from graphmb.evaluate import compute_clusters_and_stats, eval_epoch
from graphmb.utils import get_cluster_mask
from graphmb.gnn_models import name_to_model


def run_model_gnn(dataset, args, logger, nrun):
    set_seed(args.seed)
    node_names = np.array(dataset.node_names)
    RESULT_EVERY = args.evalepochs
    hidden_gnn = args.hidden_gnn
    output_dim_gnn = args.embsize_gnn
    epochs = args.epoch
    lr_gnn = args.lr_gnn
    nlayers_gnn = args.layers_gnn
    gname = args.model_name
    gmodel_type = name_to_model[gname.split("_")[0].upper()]
    clustering = args.clusteringalgo
    k = args.kclusters
    use_disconnected = not args.quick
    cluster_markers_only = args.quick
    use_edge_weights = True
    concat_features = True # bypass args, otherwise results are bad
    target_metric = args.targetmetric
    
    with mlflow.start_run(run_name=args.assembly.split("/")[-1] + "-" + args.outname):
        mlflow.log_params(vars(args))
        current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        train_log_dir = os.path.join(args.outdir, 'logs/' + args.outname + current_time + '/train')
        summary_writer = tf.summary.create_file_writer(train_log_dir)
        
        tb_handler = TensorboardLogger(summary_writer, runname=args.outname + current_time)
        logger.addHandler(tb_handler)
        tf.config.experimental_run_functions_eagerly(True)

        X, adj, cluster_mask, neg_pair_idx, pos_pair_idx = prepare_data_for_gnn(
                dataset, use_edge_weights, cluster_markers_only, use_raw=args.rawfeatures,
                binarize=args.binarize, remove_edges=args.noedges)
        if nrun == 0:
            print("logging to tensorboard")
            logger.info("******* Running model: {} **********".format(gname))
            logger.info("***** using edge weights: {} ******".format(use_edge_weights))
            logger.info("***** using disconnected: {} ******".format(use_disconnected))
            logger.info("***** concat features: {} *****".format(concat_features))
            logger.info("***** cluster markers only: {} *****".format(cluster_markers_only))
            logger.info("***** threshold adj matrix: {} *****".format(args.binarize))
            logger.info("***** self edges only: {} *****".format(args.noedges))
            logger.info("***** Using raw kmer+abund features: {}".format(args.rawfeatures))
        
            logger.info("***** SCG neg pairs: {}".format(neg_pair_idx.shape))
            logger.info("***** input features dimension: {}".format(X[cluster_mask].shape[1]))
            logger.info("***** Nodes used for clustering: {}".format(X[cluster_mask].shape[0]))

        #plot edges vs initial embs
        #id_to_scg = {i: set(dataset.contig_markers.get(node_name, {}).keys()) for i, node_name in enumerate(dataset.node_names)}
        #plot_edges_sim(X, dataset.adj_matrix, id_to_scg, f"{args.outdir}/{args.outname}_pretrain_")

        # pre train clustering
        if not args.skip_preclustering and nrun == 0:
            cluster_labels, stats, _, hq_bins = compute_clusters_and_stats(
                        X[cluster_mask], node_names[cluster_mask],
                        dataset, clustering=clustering, k=k,
                        amber=(args.labels is not None and "amber" in args.labels),
                        cuda=args.cuda,
                    )
            logger.info(f">>> Pre train stats: {str(stats)}")
        else:
            stats = {"hq": 0, "epoch":0, target_metric: 0}
        
        scores = [stats]
        losses = {"total": [], "ae": [], "gnn": [], "scg": []}
        X = X.astype(np.float32)
        features = tf.constant(X)
        input_dim_gnn = X.shape[1]
    
        logger.info(f"*** Model input dim {X.shape[1]}, GNN input dim {input_dim_gnn}")

        gnn_model = gmodel_type(
            features_shape=features.shape,
            input_dim=input_dim_gnn,
            labels=None,
            adj=adj,
            embsize=output_dim_gnn,
            hidden_units=hidden_gnn,
            layers=nlayers_gnn,
            conv_last=False,
        )
        logger.info(f"*** output clustering dim {output_dim_gnn}")

        th = TH(
            features,
            gnn_model=gnn_model,
            lr=lr_gnn,
            all_different_idx=neg_pair_idx,
            all_same_idx=pos_pair_idx,
            ae_encoder=None,
            ae_decoder=None,
            latentdim=output_dim_gnn,
            gnn_weight=float(args.gnn_alpha),
            ae_weight=float(args.ae_alpha),
            scg_weight=float(args.scg_alpha),
            num_negatives=args.negatives,
            decoder_input=args.decoder_input,
        )

        #gnn_model.summary()
        if args.eval_split == 0:
            train_idx = np.arange(len(features))
            eval_idx = []
        else:
            train_idx = np.array(random.sample(list(range(len(features))), int(len(features)*(1-args.eval_split))))
            eval_idx = np.array([x for x in np.arange(len(features)) if x not in train_idx])
            logging.info(f"**** using {len(train_idx)} for training and {len(eval_idx)} for eval")
        features = np.array(features)
        pbar_epoch = tqdm(range(epochs), disable=args.quiet, position=0)
        scores = []
        best_embs = None
        best_model = th.gnn_model.get_weights()
        best_score = 0
        best_epoch = 0
        batch_size = args.batchsize
        if batch_size == 0:
            batch_size = len(train_idx)
        all_cluster_labels = []
        step = 0
        for e in pbar_epoch:
            np.random.shuffle(train_idx)
            step += 1
            loss, gnn_losses, ae_losses = th.train_unsupervised(train_idx, scg=True)
            epoch_losses = {"Total": loss.numpy(),
                            "gnn": gnn_losses["gnn_loss"].numpy(),
                             "SCG": gnn_losses["scg_loss"].numpy(),
                             #"pred": gnn_losses["pred_loss"],
                            #'GNN  LR': float(trainer.opt._decayed_lr(float)),
                            "pos": gnn_losses["pos"].numpy(),
                            "neg": gnn_losses["neg"].numpy()}
            log_to_tensorboard(summary_writer, epoch_losses, step)
            mlflow.log_metrics(epoch_losses, step=step)
            gnn_loss = epoch_losses["gnn"]
            diff_loss = epoch_losses["SCG"]

            if args.eval_split > 0:
                eval_total_loss, eval_gnn_loss, eval_diff_loss, eval_pos_loss, \
                    eval_neg_loss = th.train_unsupervised(eval_idx, training=False)
                eval_epoch_losses = {"Eval gnn loss": float(eval_gnn_loss), "Eval SCG loss": float(eval_diff_loss),
                                     "Eval GNN loss": float(eval_total_loss),
                                     "Eval pos loss": float(eval_pos_loss),
                                     "Eval neg_loss": float(eval_neg_loss)}
                log_to_tensorboard(summary_writer, eval_epoch_losses, step)
                mlflow.log_metrics(eval_epoch_losses, step=step)
            #else:
            #    eval_loss, eval_mse1, eval_mse2, eval_kld = 0, 0, 0, 0

            #gpu_mem_alloc = tf.config.experimental.get_memory_info('GPU:0')["peak"] / 1000000 if args.cuda else 0
            gpu_mem_alloc = tf.config.experimental.get_memory_usage('GPU:0') / 1000000 if args.cuda else 0
            if (e + 1) % RESULT_EVERY == 0 and e > args.evalskip:
                gnn_input_features = features
                node_new_features = th.gnn_model(gnn_input_features, None, training=False)
                node_new_features = node_new_features.numpy()
                if concat_features:
                    node_new_features = tf.concat([gnn_input_features, node_new_features], axis=1).numpy()
                best_score, best_embs, best_epoch, scores, best_model, cluster_labels = eval_epoch(logger, summary_writer, node_new_features,
                                                                    cluster_mask, th.gnn_model.get_weights(), step, args, dataset, e, scores,
                                                                    best_score, best_embs, best_epoch, best_model, target_metric=target_metric)
                
                stats = scores[-1]
                if args.quiet:
                    logger.info(f"--- EPOCH {e:d} ---")
                    scores_string = f"HQ={stats['hq']}   Best{target_metric}={round(best_score, 3)} Best Epoch={best_epoch}"
                    logger.info(f"[{gname} {nlayers_gnn}l] L={gnn_loss:.3f} D={diff_loss:.3f} {scores_string} GPU={gpu_mem_alloc:.1f}MB")
                    logger.info(str(stats))
                mlflow.log_metrics(stats, step=step)
                all_cluster_labels.append(cluster_labels)
            scores_string = f"HQ={stats['hq']}  Best{target_metric}={round(best_score, 3)}  Best Epoch={best_epoch}"
            pbar_epoch.set_description(
                f"[{args.outname} {nlayers_gnn}l] L={gnn_loss:.3f} D={diff_loss:.3f} {scores_string} GPU={gpu_mem_alloc:.1f}MB"
            )
            total_loss = gnn_loss + diff_loss
            losses["gnn"].append(gnn_loss)
            losses["scg"].append(diff_loss)
            losses["total"].append(total_loss)

        if best_embs is None:
            best_embs = node_new_features
        gnn_model.set_weights(best_model)
        node_new_features = th.gnn_model(gnn_input_features, None, training=False)
        node_new_features = node_new_features.numpy()
        if concat_features:
            node_new_features = tf.concat([gnn_input_features, node_new_features], axis=1).numpy()
        cluster_labels, stats, _, _ = compute_clusters_and_stats(
            node_new_features, node_names, dataset,
            clustering=clustering, k=k, amber=(args.labels is not None and "amber" in args.labels),
            #cuda=args.cuda,
        )
        all_cluster_labels.append(cluster_labels)
        stats["epoch"] = e
        scores.append(stats)
        # get best stats:
        target_scores = [s[target_metric] for s in scores]
        best_idx = np.argmax(target_scores)
        mlflow.log_metrics(scores[best_idx], step=step+1)
    logger.info(f">>> best epoch all contigs: {RESULT_EVERY + (best_idx*RESULT_EVERY)} : {stats} <<<")
    logger.info(f">>> best epoch: {RESULT_EVERY + (best_idx*RESULT_EVERY)} : {scores[best_idx]} <<<")
    with open(f"{args.outdir}/{args.outname}_{nrun}_best_contig2bin.tsv", "w") as f:
        f.write("@Version:0.9.0\n@SampleID:SAMPLEID\n@@SEQUENCEID\tBINID\n")
        for i in range(len(all_cluster_labels[best_idx])):
            f.write(f"{node_names[i]}\t{all_cluster_labels[best_idx][i]}\n")
    #plot edges vs final embs
    #plot_edges_sim(best_embs, dataset.adj_matrix, id_to_scg, f"{args.outdir}/{args.outname}_posttrain_")
    return best_embs, scores[best_idx]




import os
import torch
os.environ["DGLBACKEND"] = "pytorch"
import dgl
from dgl.data import DGLDataset
import networkx as nx

class DGLAssemblyDataset(DGLDataset):
    def __init__(self, assembly):
        self.assembly = assembly
        self.logger = assembly.logger
        super().__init__(name=assembly.name + "assembly_graph", save_dir=assembly.cache_dir, force_reload=False)

    def __getitem__(self, i):
        return self.graph

    def __len__(self):
        return 1

    def process(self, root=False):
        """Open GFA file to DGL format"""
        # TODO: this should receive and assembly dataset object and initialize self.graph
        # if root:
        #    root_node = G.add_node("root", length=0)
        # TODO: skip unconnected and too short/big
        self.logger.info("creating DGL graph")
        self.graph = dgl.graph(
            (self.assembly.edges_src, self.assembly.edges_dst),
            num_nodes=len(self.assembly.node_names),
        )
        self.graph.edata["weight"] = torch.tensor(self.assembly.edge_weights)

        self.logger.info("done")
        self.graph.ndata["label"] = torch.LongTensor(
            [self.assembly.labels.index(self.assembly.node_to_label[n]) if self.assembly.node_to_label[n] in self.assembly.labels else 0 for n in self.assembly.node_names]
        )

        nx_graph = self.graph.to_networkx().to_undirected()
        self.logger.info("connected components...")
        # self.connected = [c for c in sorted(nx.connected_components(nx_graph), key=len, reverse=True) if len(c) > 1]
        # breakpoint()
        self.connected = [c for c in sorted(nx.connected_components(nx_graph), key=len, reverse=True) if len(c) > 0]
        self.logger.info((len(self.connected), "connected"))
        # for group in self.connected:
        #    self.graphs.append(
        #        dgl.node_subgraph(self.graph, [self.node_names.index(c) for c in group if c in self.node_names])
        #    )

        assert len([c for comp in self.connected for c in comp]) <= len(self.assembly.node_names)

        # self.set_node_mask()

    def save(self):
        pass
        # save graphs and labels
        # save other information in python dict
        """info_path = os.path.join(self.save_path, "cache.pkl")
        print("saving graph", info_path)
        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)
        save_info(info_path, vars(self))"""

    def load(self):
        pass
        # load processed data from directory `self.save_path`
        """ info_path = os.path.join(self.save_path, "cache.pkl")
        print("loading from", info_path)
        loaded_info = load_info(info_path)
        for key in loaded_info:
            setattr(self, key, loaded_info[key])"""

    def has_cache(self):
        """# check whether there are processed data in `self.save_path`
        info_path = os.path.join(self.save_path, "cache.pkl")
        return os.path.exists(info_path)"""
        return False


class ContigsDataset(DGLDataset):
    def __init__(
        self,
        name,
        assembly_path=None,
        assembly_name="assembly.fasta",
        graph_file="assembly_graph.gfa",
        labels=None,
        save_dir=None,
        force_reload=False,
        min_contig=1000,
        kmer=4,
        depth=None,
        markers=None,
        load_kmer=False,
        assembly_type="flye",
    ):
        self.mode = "train"
        # self.save_dir = save_dir
        self.assembly = assembly_path
        if self.assembly is None:
            self.assembly = ""
        self.readmapping = assembly_path
        self.assembly_name = assembly_name
        self.graph_file = graph_file
        self.depth = depth
        self.markers = markers
        self.contig_names = []
        self.contig_seqs = {}
        self.read_names = []
        self.node_names = []  # contig_names + read_names
        self.nodes_len = []
        self.nodes_depths = []
        self.nodes_markers = []
        self.nodes_kmer = []
        self.graphs = []
        self.edges_src = []
        self.edges_dst = []
        self.edges_weight = []
        self.nodes_data = []
        self.node_to_label = {}
        self.node_labels = []
        self.kmer = kmer
        self.load_kmer = load_kmer
        self.assembly_type = assembly_type
        if self.load_kmer:
            self.kmer_to_ids, self.canonical_k = get_kmer_to_id(self.kmer)

        self.connected = []
        self.min_contig_len = min_contig
        if labels is None:
            self.species = ["NA"]
            self.add_new_species = True
        else:
            self.species = labels
            self.add_new_species = False
        super().__init__(name=name, save_dir=save_dir, force_reload=force_reload)

    def filter_edges(self, weight=0):
        """Filter edges based on weight"""
        # print(max(self.edges_weight), min(self.edges_weight))
        if weight < 0:
            weight = sum(self.edges_weight) / len(self.edges_weight)
        # for i in range(max(self.edges_weight)):
        #    print(i, self.edges_weight.count(i))
        keep_idx = self.graph.edata["weight"] >= weight
        idx_to_remove = [i for i, x in enumerate(keep_idx) if not x]
        self.edges_src = [self.edges_src[i] for i in keep_idx]
        self.edges_dst = [self.edges_dst[i] for i in keep_idx]
        # self.graph.edata["weight"] = self.graph.edata["weight"][keep_idx]
        self.graph.remove_edges(idx_to_remove)

    def filter_contigs(self):
        # remove disconnected
        keep_idx = [i for i, c in enumerate(self.contig_names) if c in self.edges_src or c in self.edges_dst]
        # keep_idx = [i for i, c in enumerate(self.contig_names)]
        self.contig_names = [self.contig_names[i] for i in keep_idx]
        self.nodes_kmer = [self.nodes_kmer[i] for i in keep_idx]
        self.nodes_depths = [self.nodes_depths[i] for i in keep_idx]
        self.nodes_len = [self.nodes_len[i] for i in keep_idx]

    def remove_nodes(self, remove_list):
        self.graph.remove_nodes(torch.tensor(remove_list))
        # self.contig_seqs = {}
        self.node_names = [self.node_names[i] for i in range(len(self.node_names)) if i not in remove_list]
        self.nodes_len = [self.nodes_len[i] for i in range(len(self.nodes_len)) if i not in remove_list]
        self.nodes_depths = [self.nodes_depths[i] for i in range(len(self.nodes_depths)) if i not in remove_list]
        self.nodes_markers = [self.nodes_markers[i] for i in range(len(self.nodes_markers)) if i not in remove_list]
        self.nodes_kmer = [self.nodes_kmer[i] for i in range(len(self.nodes_kmer)) if i not in remove_list]
        # self.edges_src = []
        # self.edges_dst = []
        # self.edges_weight = []
        self.nodes_data = [self.nodes_data[i] for i in range(len(self.nodes_data)) if i not in remove_list]
        # self.node_to_label = {}
        self.node_labels = [self.node_labels[i] for i in range(len(self.node_labels)) if i not in remove_list]

    def rename_nodes_to_index(self):
        # self.edges_src = [self.contig_names.index(i) for i in self.edges_src if i in self.contig_names]
        # self.edges_dst = [self.contig_names.index(i) for i in self.edges_dst if i in self.contig_names]
        edge_name_to_index = {n: i for i, n in enumerate(self.contig_names)}
        self.edges_src = [edge_name_to_index[n] for n in self.edges_src]
        self.edges_dst = [edge_name_to_index[n] for n in self.edges_dst]

    def set_node_mask(self):
        """Set contig nodes"""
        self.graph.ndata["contigs"] = torch.zeros(len(self.node_names), dtype=torch.bool)
        self.graph.ndata["contigs"][: len(self.contig_names)] = True

    def get_labels_from_reads(self, reads, add_new=True):
        contig_to_species = {}
        read_to_species = {}
        # contig_lens = {}
        for r in reads:
            # print(r)
            speciesname = r.split("_reads")[0]

            if speciesname not in self.species:
                if add_new:
                    self.species.append(speciesname)
                else:
                    continue
            for m in reads[r].mappings:
                if m.contigname not in contig_to_species:  # and m.contigname in contig_lens:
                    contig_to_species[m.contigname] = {}
                if speciesname not in contig_to_species[m.contigname]:
                    contig_to_species[m.contigname][speciesname] = 0
                # contig_to_species[m.contigname][speciesname] += m.mapq  # weight mapping by quality
                contig_to_species[m.contigname][speciesname] += 1  # weight mapping by len
                read_to_species[r] = speciesname
                # reads_count = int(values[0])
            # contig_to_species[values[1]][speciesname] = reads_count
        return contig_to_species, read_to_species



__doc__ = """Iterative medoid clustering.

Usage:
>>> clusters = list(ClusterIterator(matrix))

Implements one core function, cluster, along with the helper
functions write_clusters and read_clusters.
"""

import random as _random
import numpy as _np
import torch as _torch
from collections import defaultdict as _defaultdict, deque as _deque
from math import ceil as _ceil

_DEFAULT_RADIUS = 0.06
# Distance within which to search for medoid point
_MEDOID_RADIUS = 0.05

_DELTA_X = 0.005
_XMAX = 0.3

# This is the PDF of normal with µ=0, s=0.01 from -0.075 to 0.075 with intervals
# of DELTA_X, for a total of 31 values. We multiply by _DELTA_X so the density
# of one point sums to approximately one
_NORMALPDF = _DELTA_X * _torch.Tensor(
      [2.43432053e-11, 9.13472041e-10, 2.66955661e-08, 6.07588285e-07,
       1.07697600e-05, 1.48671951e-04, 1.59837411e-03, 1.33830226e-02,
       8.72682695e-02, 4.43184841e-01, 1.75283005e+00, 5.39909665e+00,
       1.29517596e+01, 2.41970725e+01, 3.52065327e+01, 3.98942280e+01,
       3.52065327e+01, 2.41970725e+01, 1.29517596e+01, 5.39909665e+00,
       1.75283005e+00, 4.43184841e-01, 8.72682695e-02, 1.33830226e-02,
       1.59837411e-03, 1.48671951e-04, 1.07697600e-05, 6.07588285e-07,
       2.66955661e-08, 9.13472041e-10, 2.43432053e-11])

class Cluster:
    __slots__ = ['medoid', 'seed', 'members', 'pvr', 'radius', 'isdefault', 'successes', 'attempts']

    def __init__(self, medoid, seed, members, pvr, radius, isdefault, successes, attempts):
        self.medoid = medoid
        self.seed = seed
        self.members = members
        self.pvr = pvr
        self.radius = radius
        self.isdefault = isdefault
        self.successes = successes
        self.attempts = attempts

    def __repr__(self):
        return '<Cluster of medoid {}, {} members>'.format(self.medoid, len(self.members))

    def as_tuple(self, labels=None):
        if labels is None:
            return (self.medoid, {i for i in self.members})
        else:
            return (labels[self.medoid], {labels[i] for i in self.members})

    def dump(self):
        return '{}\t{}\t{}\t{}\t{}\t{}\t{}\t{}'.format(self.medoid, self.seed, self.pvr, self.radius,
        self.isdefault, self.successes, self.attempts, ','.join([str(i) for i in self.members]))

    def __str__(self):
        radius = "{:.3f}".format(self.radius)
        if self.isdefault:
            radius += " (fallback)"

        return """Cluster of medoid {}
  N members: {}
  seed:      {}
  radius:    {}
  successes: {} / {}
  pvr:       {:.1f}
  """.format(self.medoid, len(self.members), self.seed, radius, self.successes, self.attempts, self.pvr)

class ClusterGenerator:
    """Iterative medoid cluster generator. Iterate this object to get clusters.

    Inputs:
        matrix: A (obs x features) Numpy matrix of data type numpy.float32
        maxsteps: Stop searching for optimal medoid after N futile attempts [25]
        windowsize: Length of window to count successes [200]
        minsuccesses: Minimum acceptable number of successes [15]
        destroy: Save memory by destroying matrix while clustering [False]
        normalized: Matrix is already preprocessed [False]
        cuda: Accelerate clustering with GPU [False]
    """

    __slots__ = ['MAXSTEPS', 'MINSUCCESSES', 'CUDA', 'RNG', 'matrix', 'indices',
                 'seed', 'nclusters', 'peak_valley_ratio', 'attempts', 'successes',
                 'histogram', 'kept_mask']

    def __repr__(self):
        return "ClusterGenerator({} points, {} clusters)".format(len(self.matrix), self.nclusters)

    def __str__(self):
        return """ClusterGenerator({} points, {} clusters)
  CUDA:         {}
  MAXSTEPS:     {}
  MINSUCCESSES: {}
  pvr:          {}
  successes:    {}/{}
""".format(len(self.matrix), self.nclusters, self.CUDA, self.MAXSTEPS, self.MINSUCCESSES,
    self.peak_valley_ratio, self.successes, len(self.attempts))

    def _check_params(self, matrix, maxsteps, windowsize, minsuccesses):
        """Checks matrix, and maxsteps."""

        if matrix.dtype != _np.float32:
            raise(ValueError("Matrix must be of dtype float32"))

        if maxsteps < 1:
            raise ValueError('maxsteps must be a positive integer, not {}'.format(maxsteps))

        if windowsize < 1:
            raise ValueError('windowsize must be at least 1, not {}'.format(windowsize))

        if minsuccesses < 1 or minsuccesses > windowsize:
            raise ValueError('minsuccesses must be between 1 and windowsize, not {}'.format(minsuccesses))

        if len(matrix) < 1:
            raise ValueError('Matrix must have at least 1 observation.')

    def _init_histogram_kept_mask(self, N):
        "N is number of contigs"
        if _torch.__version__ >= '1.2':
            # https://github.com/pytorch/pytorch/issues/20208 fixed in PyTorch 1.2
            cuda_hist_dtype = _torch.float
            # Masking using uint8 tensors deprecated in PyTorch 1.2
            kept_mask_dtype = _torch.bool
        else:
            cuda_hist_dtype = _torch.long
            kept_mask_dtype = _torch.uint8

        kept_mask = _torch.ones(N, dtype=kept_mask_dtype)

        if self.CUDA:
            histogram = _torch.empty(_ceil(_XMAX/_DELTA_X), dtype=cuda_hist_dtype).cuda()
            kept_mask = kept_mask.cuda()
        else:
            histogram = _torch.empty(_ceil(_XMAX/_DELTA_X))

        return histogram, kept_mask

    def __init__(self, matrix, maxsteps=25, windowsize=200, minsuccesses=20, destroy=False,
                normalized=False, cuda=False):
        self._check_params(matrix, maxsteps, windowsize, minsuccesses)
        if not destroy:
            matrix = matrix.copy()

        # Shuffle matrix in unison to prevent seed sampling bias. Indices keeps
        # track of which points are which.
        _np.random.RandomState(0).shuffle(matrix)
        indices = _np.random.RandomState(0).permutation(len(matrix))
        indices = _torch.from_numpy(indices)
        matrix = _torch.from_numpy(matrix)

        if not normalized:
            _normalize(matrix, inplace=True)

        # Move to GPU
        if cuda:
            matrix = matrix.cuda()

        self.MAXSTEPS = maxsteps
        self.MINSUCCESSES = minsuccesses
        self.CUDA = cuda
        self.RNG = _random.Random(0)

        self.matrix = matrix
        # This refers to the indices of the original matrix. As we remove points, these
        # indices do not correspond to merely range(len(matrix)) anymore.
        self.indices = indices
        self.seed = -1
        self.nclusters = 0
        self.peak_valley_ratio = 0.1
        self.attempts = _deque(maxlen=windowsize)
        self.successes = 0

        histogram, kept_mask = self._init_histogram_kept_mask(len(indices))
        self.histogram = histogram
        self.kept_mask = kept_mask

    # It's an iterator itself
    def __iter__(self):
        return self

    def __next__(self):
        # Stop criterion. For CUDA, inplace masking the array is too slow, so the matrix is
        # unchanged. On CPU, we continually modify the matrix by removing rows.
        if self.CUDA:
            if not _torch.any(self.kept_mask).item():
                raise StopIteration
        elif len(self.matrix) == 0:
            raise StopIteration

        cluster, medoid, points = self._findcluster()
        self.nclusters += 1

        for point in points:
            self.kept_mask[point] = 0

        # Remove all points that's been clustered away. Is slow it itself, but speeds up
        # distance calculation by having fewer points. Worth it on CPU, not on GPU
        if not self.CUDA:
            self.matrix = self.matrix[self.kept_mask]
            #_vambtools.torch_inplace_maskarray(self.matrix, self.kept_mask)
            self.indices = self.indices[self.kept_mask] # no need to inplace mask small array
            self.kept_mask.resize_(len(self.matrix))
            self.kept_mask[:] = 1

        return cluster

    def _findcluster(self):
        """Finds a cluster to output."""
        threshold = None

        # Keep looping until we find a cluster
        while threshold is None:
            # If on GPU, we need to take next seed which has not already been clusted out.
            # if not, clustered points have been removed, so we can just take next seed
            if self.CUDA:
                self.seed = (self.seed + 1) % len(self.matrix)
                while self.kept_mask[self.seed] == False:
                    self.seed = (self.seed + 1) % len(self.matrix)
            else:
                self.seed = (self.seed + 1) % len(self.matrix)

            medoid, distances = _wander_medoid(self.matrix, self.kept_mask, self.seed, self.MAXSTEPS, self.RNG, self.CUDA)

            # We need to make a histogram of only the unclustered distances - when run on GPU
            # these have not been removed and we must use the kept_mask
            if self.CUDA:
                _torch.histc(distances[self.kept_mask], len(self.histogram), 0, _XMAX, out=self.histogram)
            else:
                _torch.histc(distances, len(self.histogram), 0, _XMAX, out=self.histogram)
            self.histogram[0] -= 1 # Remove distance to self

            threshold, success = _find_threshold(self.histogram, self.peak_valley_ratio, self.CUDA)

            # If success is not None, either threshold detection failed or succeded.
            if success is not None:
                # Keep accurately track of successes if we exceed maxlen
                if len(self.attempts) == self.attempts.maxlen:
                    self.successes -= self.attempts.popleft()

                # Add the current success to count
                self.successes += success
                self.attempts.append(success)

                # If less than minsuccesses of the last maxlen attempts were successful,
                # we relax the clustering criteria and reset counting successes.
                if len(self.attempts) == self.attempts.maxlen and self.successes < self.MINSUCCESSES:
                    self.peak_valley_ratio += 0.1
                    self.attempts.clear()
                    self.successes = 0

        # These are the points of the final cluster AFTER establishing the threshold used
        points = _smaller_indices(distances, self.kept_mask, threshold, self.CUDA)
        isdefault = success is None and threshold == _DEFAULT_RADIUS and self.peak_valley_ratio > 0.55

        cluster = Cluster(self.indices[medoid].item(), self.seed, self.indices[points].numpy(),
                        self.peak_valley_ratio,
                          threshold, isdefault, self.successes, len(self.attempts))
        return cluster, medoid, points

def _calc_densities(histogram, cuda, pdf=_NORMALPDF):
    """Given an array of histogram, smoothes the histogram."""
    pdf_len = len(pdf)

    if cuda:
        histogram = histogram.cpu()

    densities = _torch.zeros(len(histogram) + pdf_len - 1)
    for i in range(len(densities) - pdf_len + 1):
        densities[i:i+pdf_len] += pdf * histogram[i]

    densities = densities[15:-15]

    return densities

def _find_threshold(histogram, peak_valley_ratio, cuda):
    """Find a threshold distance, where where is a dip in point density
    that separates an initial peak in densities from the larger bulk around 0.5.
    Returns (threshold, success), where succes is False if no threshold could
    be found, True if a good threshold could be found, and None if the point is
    alone, or the threshold has been used.
    """
    peak_density = 0
    peak_over = False
    minimum_x = None
    density_at_minimum = None
    threshold = None
    success = False
    delta_x = _XMAX / len(histogram)

    # If the point is a loner, immediately return a threshold in where only
    # that point is contained.
    if histogram[:10].sum().item() == 0:
        return 0.025, None

    densities = _calc_densities(histogram, cuda)

    # Else we analyze the point densities to find the valley
    x = 0
    for density in densities:
        # Define the first "peak" in point density. That's simply the max until
        # the peak is defined as being over.
        if not peak_over and density > peak_density:
            # Do not accept first peak to be after x = 0.1
            if x > 0.1:
                break
            peak_density = density

        # Peak is over when density drops below 60% of peak density
        if not peak_over and density < 0.6 * peak_density:
            peak_over = True
            density_at_minimum = density

        # If another peak is detected, we stop
        if peak_over and density > 1.5 * density_at_minimum:
            break

        # Now find the minimum after the peak
        if peak_over and density < density_at_minimum:
            minimum_x, density_at_minimum = x, density

            # If this minimum is below ratio * peak, it's accepted as threshold
            if density < peak_valley_ratio * peak_density:
                threshold = minimum_x
                success = True

        x += delta_x

    # Don't allow a threshold too high - this is relaxed with p_v_ratio
    if threshold is not None and threshold > 0.2 + peak_valley_ratio:
        threshold = None
        success = False

    # If ratio has been set to 0.6, we do not accept returning no threshold.
    if threshold is None and peak_valley_ratio > 0.55:
        threshold = _DEFAULT_RADIUS
        success = None

    return threshold, success

def _smaller_indices(tensor, kept_mask, threshold, cuda):
    """Get all indices where the tensor is smaller than the threshold.
    Uses Numpy because Torch is slow - See https://github.com/pytorch/pytorch/pull/15190"""

    # If it's on GPU, we remove the already clustered points at this step.
    if cuda:
        return _torch.nonzero((tensor <= threshold) & kept_mask).flatten()
    else:
        arr = tensor.numpy()
        indices = (arr <= threshold).nonzero()[0]
        torch_indices = _torch.from_numpy(indices)
        return torch_indices

def _normalize(matrix, inplace=False):
    """Preprocess the matrix to make distance calculations faster.
    The distance functions in this module assumes input has been normalized
    and will not work otherwise.
    """
    if isinstance(matrix, _np.ndarray):
        matrix = _torch.from_numpy(matrix)

    if not inplace:
        matrix = matrix.clone()

    # If any rows are kept all zeros, the distance function will return 0.5 to all points
    # inclusive itself, which can break the code in this module
    zeromask = matrix.sum(dim=1) == 0
    matrix[zeromask] = 1/matrix.shape[1]
    matrix /= (matrix.norm(dim=1).reshape(-1, 1) * (2 ** 0.5))
    return matrix


def _calc_distances(matrix, index):
    "Return vector of cosine distances from rows of normalized matrix to given row."
    dists = 0.5 - matrix.matmul(matrix[index])
    dists[index] = 0.0 # avoid float rounding errors
    return dists

def _sample_medoid(matrix, kept_mask, medoid, threshold, cuda):
    """Returns:
    - A vector of indices to points within threshold
    - A vector of distances to all points
    - The mean distance from medoid to the other points in the first vector
    """

    distances = _calc_distances(matrix, medoid)
    cluster = _smaller_indices(distances, kept_mask, threshold, cuda)

    if len(cluster) == 1:
        average_distance = 0.0
    else:
        average_distance = distances[cluster].sum().item() / (len(cluster) - 1)

    return cluster, distances, average_distance


def _wander_medoid(matrix, kept_mask, medoid, max_attempts, rng, cuda):
    """Keeps sampling new points within the cluster until it has sampled
    max_attempts without getting a new set of cluster with lower average
    distance"""

    futile_attempts = 0
    tried = {medoid} # keep track of already-tried medoids
    cluster, distances, average_distance = _sample_medoid(matrix, kept_mask, medoid, _MEDOID_RADIUS, cuda)

    while len(cluster) - len(tried) > 0 and futile_attempts < max_attempts:
        sampled_medoid = rng.choice(cluster).item()

         # Prevent sampling same medoid multiple times.
        while sampled_medoid in tried:
            sampled_medoid = rng.choice(cluster).item()

        tried.add(sampled_medoid)

        sampling = _sample_medoid(matrix, kept_mask, sampled_medoid, _MEDOID_RADIUS, cuda)
        sample_cluster, sample_distances, sample_avg = sampling

        # If the mean distance of inner points of the sample is lower,
        # we move the medoid and reset the futile_attempts count
        if sample_avg < average_distance:
            medoid = sampled_medoid
            cluster = sample_cluster
            average_distance = sample_avg
            futile_attempts = 0
            tried = {medoid}
            distances = sample_distances

        else:
            futile_attempts += 1

    return medoid, distances

def cluster(matrix, labels=None, maxsteps=25, windowsize=200, minsuccesses=20, destroy=False,
            normalized=False, cuda=False):
    """Create iterable of (medoid, {point1, point2 ...}) tuples for each cluster.

    Inputs:
        matrix: A (obs x features) Numpy matrix of data type numpy.float32
        labels: None or list of labels of points [None = range(len(matrix))]
        maxsteps: Stop searching for optimal medoid after N futile attempts [25]
        windowsize: Length of window to count successes [200]
        minsuccesses: Minimum acceptable number of successes [15]
        destroy: Save memory by destroying matrix while clustering [False]
        normalized: Matrix is already preprocessed [False]
        cuda: Accelerate clustering with GPU [False]

    Output: Generator of (medoid, {point1, point2 ...}) tuples for each cluster.
    """
    if labels is not None and len(matrix) != len(labels):
        raise ValueError("Got {} labels for {} points".format(len(labels), len(matrix)))

    it = ClusterGenerator(matrix, maxsteps, windowsize, minsuccesses, destroy, normalized, cuda)
    for cluster in it:
        yield cluster.as_tuple(labels)

def pairs(clustergenerator, labels):
    """Create an iterable of (N, {label1, label2 ...}) for each
    cluster in a ClusterGenerator, where N is "1", "2", "3", etc.
    Useful to pass to e.g. vambtools.writer_clusters.

    Inputs:
        clustergenerator: A ClusterGenerator object
        labels: List or array of cluster labels
    Output:
        Generator yielding ("1", {label1, label2 ... }) for each cluster
    """
    maxindex = clustergenerator.indices.max()
    if len(labels) < maxindex:
        raise ValueError("Cluster generator contains point no {}, "
                         "but was given only {} labels".format(maxindex, len(labels)))

    return ((str(i+1), c.as_tuple(labels)[1]) for (i, c) in enumerate(clustergenerator))


import tensorflow as tf

from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, Activation
from tensorflow.keras.layers import BatchNormalization, Lambda
from tensorflow.keras.layers import Dropout, Layer

def laf_init_uniform(shape, dtype=None):
    a = tf.random.uniform(shape ,0,1, dtype=dtype)
    return a

def laf_init_normal(shape, dtype=None):
    b = tf.random.normal(shape, 1.0, 0.1, dtype=dtype)
    return b

class BiasLayer(Layer):
    def __init__(self, *args, **kwargs):
        super(BiasLayer, self).__init__(*args, **kwargs)

    def build(self, input_shape):
        self.bias = self.add_weight('bias',
                                    shape=input_shape[1:],
                                    initializer='zeros',
                                    trainable=True)
    def call(self, x):
        return x + self.bias

class LAF(Layer):
    def __init__(self, units=4, eps=1e-7):
        super(LAF, self).__init__()
        self.units = units
        self.eps = eps
        
    def build(self, input_shape):

        self.b = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='b')
        self.d = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='d')
        self.f = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='f')
        self.h = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='h')

        self.a = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='a')
        self.c = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='c')
        self.e = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='e')
        self.g = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_uniform,
                                 trainable=True, name='g')
       
        self.alpha = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_normal,
                                 trainable=True, name='alpha')
        self.beta  = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_normal,
                                 trainable=True, name='beta')
        self.gamma = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_normal,
                                 trainable=True, name='gamma')
        self.delta = self.add_weight(shape=(1, self.units, 1),
                                 initializer=laf_init_normal,
                                 trainable=True, name='delta')
    def call(self, inputs):
        eps = 1e-6
        adj, inputs = inputs

        x = inputs
        x  = tf.reshape(x, [-1, self.units, inputs.shape[-1]//self.units])
        sig_x  = tf.clip_by_value(tf.nn.sigmoid(x), eps, 1-eps)
        n_sig_x = 1.0 - sig_x

        x_b = tf.math.pow(sig_x, tf.nn.relu(self.b))
        x_d = tf.math.pow(n_sig_x, tf.nn.relu(self.d))
        x_f = tf.math.pow(sig_x, tf.nn.relu(self.f))
        x_h = tf.math.pow(n_sig_x, tf.nn.relu(self.h))

        x_b = tf.reshape(x_b, [-1, inputs.shape[-1]])
        x_d = tf.reshape(x_d, [-1, inputs.shape[-1]])
        x_f = tf.reshape(x_f, [-1, inputs.shape[-1]])
        x_h = tf.reshape(x_h, [-1, inputs.shape[-1]])

        x_b = tf.sparse.sparse_dense_matmul(adj, x_b) + eps
        x_d = tf.sparse.sparse_dense_matmul(adj, x_d) + eps
        x_f = tf.sparse.sparse_dense_matmul(adj, x_f) + eps
        x_h = tf.sparse.sparse_dense_matmul(adj, x_h) + eps

        x_b = tf.reshape(x_b, [-1, self.units, inputs.shape[-1]//self.units])
        x_d = tf.reshape(x_d, [-1, self.units, inputs.shape[-1]//self.units])
        x_f = tf.reshape(x_f, [-1, self.units, inputs.shape[-1]//self.units])
        x_h = tf.reshape(x_h, [-1, self.units, inputs.shape[-1]//self.units])

        x_ab = tf.math.pow(x_b,tf.nn.relu(self.a) ) * self.alpha
        x_cd = tf.math.pow(x_d,tf.nn.relu(self.c) ) * self.beta
        x_ef = tf.math.pow(x_f,tf.nn.relu(self.e) ) * self.gamma
        x_gh = tf.math.pow(x_h,tf.nn.relu(self.g) ) * self.delta

        den = (x_ef + x_gh)
        num = (x_ab + x_cd) * den

        out = num / (den * den + 1e-3)
        out = tf.reshape(out, (-1, inputs.shape[-1]))
        return out
    
    
# https://github.com/danielegrattarola/keras-gat/blob/master/keras_gat/graph_attention_layer.py
# https://www.programmersought.com/article/38174813555/

class GraphAttention(Layer):
    def __init__(self,
                 F_,
                 attn_heads=1,
                 attn_heads_reduction='concat',  # {'concat', 'average'}
                 dropout_rate=0.5,
                 activation='relu',
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 attn_kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 attn_kernel_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 attn_kernel_constraint=None,
                 laf_units=None,
                 **kwargs):
        if attn_heads_reduction not in {'concat', 'average'}:
            raise ValueError('Possbile reduction methods: concat, average')

        self.F_ = F_  # Number of output features (F' in the paper)
        self.attn_heads = attn_heads  # Number of attention heads (K in the paper)
        self.attn_heads_reduction = attn_heads_reduction  # Eq. 5 and 6 in the paper
        self.dropout_rate = dropout_rate  # Internal dropout rate
        self.activation = tf.keras.activations.get(activation)  # Eq. 4 in the paper
        self.use_bias = use_bias

        self.kernel_initializer = tf.keras.initializers.get(kernel_initializer)
        self.bias_initializer = tf.keras.initializers.get(bias_initializer)
        self.attn_kernel_initializer = tf.keras.initializers.get(attn_kernel_initializer)

        self.kernel_regularizer = tf.keras.regularizers.get(kernel_regularizer)
        self.bias_regularizer = tf.keras.regularizers.get(bias_regularizer)
        self.attn_kernel_regularizer = tf.keras.regularizers.get(attn_kernel_regularizer)
        self.activity_regularizer = tf.keras.regularizers.get(activity_regularizer)

        self.kernel_constraint = tf.keras.constraints.get(kernel_constraint)
        self.bias_constraint = tf.keras.constraints.get(bias_constraint)
        self.attn_kernel_constraint = tf.keras.constraints.get(attn_kernel_constraint)
        self.supports_masking = False

        # Populated by build()
        self.kernels = []       # Layer kernels for attention heads
        self.biases = []        # Layer biases for attention heads
        self.attn_kernels = []  # Attention kernels for attention heads

        if attn_heads_reduction == 'concat':
            # Output will have shape (..., K * F')
            self.output_dim = self.F_ * self.attn_heads
        else:
            # Output will have shape (..., F')
            self.output_dim = self.F_

        self.laf_units = laf_units
        if self.laf_units is not None:
            self.laf_agg = LAF(laf_units)

        super(GraphAttention, self).__init__(**kwargs)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        F = input_shape[1][-1]

        # Initialize weights for each attention head
        for head in range(self.attn_heads):
            # Layer kernel
            kernel = self.add_weight(shape=(F, self.F_),
                                     initializer=self.kernel_initializer,
                                     regularizer=self.kernel_regularizer,
                                     constraint=self.kernel_constraint,
                                     name='kernel_{}'.format(head))
            self.kernels.append(kernel)

            # # Layer bias
            if self.use_bias:
                bias = self.add_weight(shape=(self.F_, ),
                                       initializer=self.bias_initializer,
                                       regularizer=self.bias_regularizer,
                                       constraint=self.bias_constraint,
                                       name='bias_{}'.format(head))
                self.biases.append(bias)

            # Attention kernels
            attn_kernel_self = self.add_weight(shape=(self.F_, 1),
                                               initializer=self.attn_kernel_initializer,
                                               regularizer=self.attn_kernel_regularizer,
                                               constraint=self.attn_kernel_constraint,
                                               name='attn_kernel_self_{}'.format(head),)
            attn_kernel_neighs = self.add_weight(shape=(self.F_, 1),
                                                 initializer=self.attn_kernel_initializer,
                                                 regularizer=self.attn_kernel_regularizer,
                                                 constraint=self.attn_kernel_constraint,
                                                 name='attn_kernel_neigh_{}'.format(head))
            self.attn_kernels.append([attn_kernel_self, attn_kernel_neighs])
        self.built = True

    def call(self, inputs, training=True):
        A = inputs[0]  # Adjacency matrix (N x N) sparse
        X = inputs[1]  # Node features (N x F)

        outputs = []
        for head in range(self.attn_heads):
            kernel = self.kernels[head]  # W in the paper (F x F')
            attention_kernel = self.attn_kernels[head]  # Attention kernel a in the paper (2F' x 1)

            # Compute inputs to attention network
            features = tf.matmul(X, kernel)  # (N x F')

            # Compute feature combinations
            # Note: [[a_1], [a_2]]^T [[Wh_i], [Wh_2]] = [a_1]^T [Wh_i] + [a_2]^T [Wh_j]
            attn_for_self   = tf.matmul(features, attention_kernel[0])  # (N x 1), [a_1]^T [Wh_i]
            attn_for_neighs = tf.matmul(features, attention_kernel[1])  # (N x 1), [a_2]^T [Wh_j]

            # Attention head a(Wh_i, Wh_j) = a^T [[Wh_i], [Wh_j]]
            con_sa_1 = A * attn_for_self 
            con_sa_2 = A * tf.transpose(attn_for_neighs, [1,0])
            
            weights = tf.sparse.add(con_sa_1, con_sa_2)  # concatenation
            weights_act = tf.SparseTensor(indices=weights.indices,
                                          values=tf.nn.leaky_relu(weights.values, alpha=0.2),
                                          dense_shape=weights.dense_shape)
            
            attention = tf.sparse.softmax(weights_act)
            if training and self.dropout_rate > 0.0:
                attention = tf.SparseTensor(indices=attention.indices,
                                            values=tf.nn.dropout(attention.values, self.dropout_rate),
                                            dense_shape=attention.dense_shape)
            if training and self.dropout_rate > 0.0:
                features = tf.nn.dropout(features, self.dropout_rate)
            
            if self.laf_units is not None:
                node_features = self.laf_agg((attention, features))
            else:
                node_features = tf.sparse.sparse_dense_matmul(attention, features)
            
            if self.use_bias:
                node_features = node_features + self.biases[head]
                
            # Add output of attention head to final output
            outputs.append(node_features)

        # Aggregate the heads' output according to the reduction method
        output = tf.concat(outputs, axis=1)  # (N x KF')
        if self.attn_heads_reduction != 'concat':
            output = tf.reshape(output, [-1, self.attn_heads, self.F_])
            output = tf.reduce_mean(output, axis=-2)

        output = self.activation(output)
        return output

    def compute_output_shape(self, input_shape):
        output_shape = input_shape[0][0], self.output_dim
        return output_shape



from glob import escape
from importlib.resources import files
import itertools
import random
import os
import pickle
import pdb
import shutil
from typing import Dict, List, Tuple, Sequence
from collections import Counter
import numpy as np
import networkx as nx
import scipy.stats as stats
import scipy.sparse
from graphmb.evaluate import (
    read_marker_gene_sets,
    read_contig_genes,
    get_markers_to_contigs,
)


#BACTERIA_MARKERS = "data/Bacteria.ms"
#kernel = np.load("data/kernel.npz")['arr_0']
BACTERIA_MARKERS = files('graphmb.data').joinpath('Bacteria.ms')
kernel = np.load(files('graphmb.data').joinpath('kernel.npz'))['arr_0']

def count_kmers(seq, k, kmer_to_id, canonical_k):
    # Used in case kmers are used as input features
    # https://stackoverflow.com/q/22428020
    #breakpoint()
    kmer_letters = set(["A", "T", "C", "G"])
    kmers = [seq[i : i + k] for i in range(len(seq) - k + 1) if set(seq[i : i + k]).issubset(kmer_letters)]
    kmers = [kmer_to_id[k] for k in kmers]
    kmer_counts = Counter(kmers)
    counts = np.array([kmer_counts[k] for k in range(canonical_k)])
    counts = counts / counts.sum()
    counts += -(1/(4**k))
    counts = np.dot(counts, kernel)
    return counts

def get_kmer_to_id(kmer, combine_revcomp=False):
    kmer_to_ids = {}
    BASE_COMPLEMENT = {"A": "T", "T": "A", "G": "C", "C": "G"}
    all_kmers = itertools.product("ACGT", repeat=kmer)
    all_kmers = ["".join(k) for k in all_kmers]
    new_id = 0
    for kmer in all_kmers:
        if kmer not in kmer_to_ids:
            kmer_to_ids[kmer] = new_id
            if combine_revcomp:
                rev_compl = "".join(tuple([BASE_COMPLEMENT[x] for x in reversed(kmer)]))
                kmer_to_ids[rev_compl] = new_id
            new_id += 1
    return kmer_to_ids, new_id

def process_node_name(name: str, assembly_type: str="flye") -> str:
    contig_name = name.strip().split(" ")[0]
    #if assembly_type == "spades":
    contig_name = "_".join(contig_name.split("_")[:2])
    return contig_name


class AssemblyDataset:
    # Read contig names from fasta file and save to disk
    # read graph
    # other features are loaded by functions
    # everything as numpy to make it easier to convert
    def __init__(
        self,
        name,
        logger,
        data_dir,
        fastafile,
        graphfile,
        depthfile,
        scgfile,
        labelsfile,
        featuresfile,
        cache_dir,
        assemblytype="flye",
        min_contig_length=0,
        contignodes=False
    ):
        self.name = name
        self.logger = logger
        self.data_dir = data_dir
        self.fastafile = fastafile
        self.graphfile = graphfile
        self.depthfile = depthfile
        self.assembly_type = assemblytype
        self.cache_dir = cache_dir
        self.labelsfile = labelsfile
        self.featuresfile = featuresfile
        self.scgfile = scgfile
        self.load_kmer = True
        self.kmer = 4
        self.min_contig_length = min_contig_length
        self.contignodes = contignodes
        if self.load_kmer:
            self.kmer_to_ids, self.canonical_k = get_kmer_to_id(self.kmer)

        # initialize data features which can be cached as numpy or pickle
        self.node_names = []
        self.node_seqs = {}  # should it be saved?
        self.graph_nodes = []  # nodes that are part of the assembly graph
        self.graph_paths = {} # contig paths identified by assembler
        self.node_lengths = []
        self.node_kmers = []
        self.node_depths = []
        self.edges_src = []
        self.edges_dst = []
        self.edge_weights = []
        self.adj_matrix = None
        self.neg_pairs_idx = None  # cached on the main function, only necessary for TF models

        # initialize labels
        self.labels = []
        self.node_to_label = {}
        self.label_to_node = {}
        
        self.contig_markers = {}
        self.ref_marker_sets = {}

    def read_assembly(self):
        """Read assembly files, convert to numpy arrays and save to disk"""
        self.logger.info("processing sequences {}".format(os.path.join(self.data_dir, self.fastafile)))
        self.read_seqs()
        self.logger.info(f"read {len(self.node_seqs)} seqs")
        np.save(os.path.join(self.cache_dir, "node_names.npy"), self.node_names)
        np.save(os.path.join(self.cache_dir, "node_lengths.npy"), self.node_lengths)
        np.save(os.path.join(self.cache_dir, "node_attributes_kmer.npy"), self.node_kmers.astype(float))
        pickle.dump(self.node_seqs, open(os.path.join(self.cache_dir, "node_seqs.pkl"), "wb"))
        if os.path.exists(os.path.join(self.data_dir, self.graphfile)):
            if not self.contignodes:
                self.logger.info("processing GFA file (edge nodes) {}".format(os.path.join(self.data_dir, self.graphfile)))
                self.read_gfa()
            else:
                self.logger.info("processing GFA file (contig nodes) {}".format(os.path.join(self.data_dir, self.graphfile)))
                self.read_gfa_contigs()
            self.logger.info(f"read {len(self.edges_src)}, edges")
            np.save(os.path.join(self.cache_dir, "edge_weights.npy"), self.edge_weights)
            scipy.sparse.save_npz(os.path.join(self.cache_dir, "adj_sparse.npz"), self.adj_matrix)
            np.save(os.path.join(self.cache_dir, "graph_nodes.npy"), self.graph_nodes)
            pickle.dump(self.graph_paths, open(os.path.join(self.cache_dir, "graph_paths.pkl"), 'wb'))
        # self.filter_contigs()
        # self.rename_nodes_to_index()
        # self.nodes_depths = np.array(self.nodes_depths)
        # read lengths
        self.logger.info("reading depths")
        self.read_depths()
        np.save(os.path.join(self.cache_dir, "node_attributes_depth.npy"), np.array(self.node_depths))
        self.logger.info("reading labels")
        self.read_labels()
        np.save(os.path.join(self.cache_dir, "node_to_label.npy"), self.node_to_label)
        np.save(os.path.join(self.cache_dir, "label_to_node.npy"), self.label_to_node)
        np.save(os.path.join(self.cache_dir, "labels.npy"), self.labels)
        self.logger.info("Saved cache to {}".format(self.cache_dir))
        # print("reading SCGs")
        # self.read_scgs()

    def print_stats(self):
        print("==============================")
        print("DATASET STATS:")
        # get:
        #   number of sequences
        #   number of contigs
        #   length of contigs (sum and average and N50)
        #   coverage samples from jgi file
        #   assembly_graph.file exists, number of edges, number of paths
        print("number of sequences: {}".format(len(self.node_names)))
        print("assembly length: {} Gb".format(round(sum(self.node_lengths) / 1000000000, 3)))
        print("assembly N50: {} Mb".format(round(self.calculate_n50()/1000000, 3)))
        print("assembly average length (Mb): {} max: {} min: {}".format(round(np.mean(self.node_lengths)/1000000, 3),
                                                                   round(np.max(self.node_lengths)/1000000, 3),
                                                                   round(np.min(self.node_lengths)/1000000, 3)))
        print("coverage samples: {}".format(len(self.node_depths[0])))
        if os.path.exists(os.path.join(self.data_dir, self.graphfile)) or \
            len(self.edges_src) > 0:
            print("Graph file found and read")
            print("graph edges: {}".format(len(self.edges_src)))
            print("contig paths: {}".format(len(self.graph_paths)))
        else:
            print("No assembly graph loaded")
        #   contigs with markers on marker_gene_stats
        #   stats with SCGs (max/min # of contigs, etc)

        if len(self.contig_markers) > 0:
            print("total ref markers sets: {}".format(len(self.ref_marker_sets)))
            print("total ref markers: {}".format(len(self.markers)))
            n_of_markers = [len(x) for x in self.contig_markers.values() if len(x) > 0]
            print("contigs with one or more markers: {}/{}".format(len(n_of_markers),
                                                                    len(self.node_names)))
            
            print("max SCGs on one contig: {}, average(excluding 0): {}".format(max(n_of_markers),
                                                            round(np.mean(n_of_markers), 3)))
            self.estimate_n_genomes()
            print("SCG contig count min: {} contigs".format(min(self.scg_counts.values())))
            self.get_edges_with_same_scgs()
        else:
            print("No SCG markers")
        # labels
        if self.labelsfile is not None or len(self.labels) > 1:
            print("number of GS labels: {}".format(len(self.labels)))
            nodes_not_na = [n for n in self.node_names if n in self.node_to_label and \
                                                     self.node_to_label[n] != "NA"]
            print("nodes with labels: {}".format(len(nodes_not_na)))
        print("==============================")
                                                          
    def calculate_n50(self):
        # https://eaton-lab.org/slides/genomics/answers/nb-4.1-numpy.html
        contig_sizes = self.node_lengths.copy()
        contig_sizes.sort()
        contig_sizes = contig_sizes[::-1]
        total_len = contig_sizes.sum()
        half_total_len = total_len / 2
        contig_sum_lens = np.zeros(contig_sizes.size, dtype=int)
        for i in range(contig_sizes.size):
            contig_sum_lens[i] = contig_sizes[i:].sum()
        which_contigs_longer_than_half = contig_sum_lens > half_total_len
        contigs_longer_than_half = contig_sizes[which_contigs_longer_than_half]
        n50 = contigs_longer_than_half.min()
        return n50

    def read_cache(self, load_graph=True):
        prefix = os.path.join(self.cache_dir, "{}")
        self.node_names = list(np.load(prefix.format("node_names.npy")))
        self.node_seqs = pickle.load(open(prefix.format("node_seqs.pkl"), "rb"))
        # graph nodes
        self.node_lengths = np.load(prefix.format("node_lengths.npy"))
        if load_graph:
            self.graph_nodes = np.load(prefix.format("graph_nodes.npy"))
            self.graph_paths = pickle.load(open(prefix.format("graph_paths.pkl"), 'rb'))
            self.adj_matrix = scipy.sparse.load_npz(prefix.format("adj_sparse.npz"))
            self.edge_weights = self.adj_matrix.data
            self.edges_src = self.adj_matrix.row
            self.edges_dst = self.adj_matrix.col
        self.node_kmers = np.load(prefix.format("node_attributes_kmer.npy"))
        self.node_depths = np.load(prefix.format("node_attributes_depth.npy"))
        # self.edge_weights = np.load(prefix.format("edge_weights.npy"))
        
        # labels
        self.node_to_label = np.load("{}/node_to_label.npy".format(self.cache_dir), allow_pickle=True)[()]
        self.label_to_node = np.load("{}/label_to_node.npy".format(self.cache_dir), allow_pickle=True)[()]
        self.labels = list(np.load("{}/labels.npy".format(self.cache_dir)))
        if os.path.exists(prefix.format("true_adj_sparse.npz")):
            self.true_adj_matrix = scipy.sparse.load_npz(prefix.format("true_adj_sparse.npz"))

    def check_cache(self, require_graph=True):
        """check if all necessary files exist in cache"""
        prefix = os.path.join(self.cache_dir, "{}")
        if require_graph:
            return (
                os.path.exists(prefix.format("node_names.npy"))
                and os.path.exists(prefix.format("node_lengths.npy"))
                and os.path.exists(prefix.format("node_attributes_kmer.npy"))
                #and os.path.exists(prefix.format("node_attributes_depth.npy"))
                and os.path.exists(prefix.format("edge_weights.npy"))
                and os.path.exists(prefix.format("adj_sparse.npz"))
                and os.path.exists(prefix.format("graph_nodes.npy"))
                and os.path.exists(prefix.format("graph_paths.pkl"))
                and os.path.exists(prefix.format("node_to_label.npy"))
                and os.path.exists(prefix.format("label_to_node.npy"))
                and os.path.exists(prefix.format("labels.npy"))
            )
        else:
            return (
                os.path.exists(prefix.format("node_names.npy"))
                and os.path.exists(prefix.format("node_lengths.npy"))
                and os.path.exists(prefix.format("node_attributes_kmer.npy"))
                #and os.path.exists(prefix.format("node_attributes_depth.npy"))
                and os.path.exists(prefix.format("node_to_label.npy"))
                and os.path.exists(prefix.format("label_to_node.npy"))
                and os.path.exists(prefix.format("labels.npy"))
            )

    def read_seqs(self):
        """Read sequences from fasta file, write to self.node_seqs"""
        node_lengths = {}
        node_kmers = {}
        contig_name = None
        with open(os.path.join(self.data_dir, self.fastafile), "r") as f:
            for line in f:
                if line.startswith(">"):
                    if len(self.node_names) > 0:
                        # finish last contig
                        node_lengths[contig_name] = len(self.node_seqs[contig_name])
                        if self.load_kmer:
                            kmers = count_kmers(
                                self.node_seqs[contig_name], self.kmer, self.kmer_to_ids, self.canonical_k
                            )
                            node_kmers[contig_name] = kmers
                    contig_name = process_node_name(line[1:], self.assembly_type)
                    self.node_names.append(contig_name)
                    self.node_seqs[contig_name] = ""
                else:
                    self.node_seqs[contig_name] += line.strip()
        # add last
        node_lengths[contig_name] = len(self.node_seqs[contig_name])
        if self.load_kmer:
            kmers = count_kmers(self.node_seqs[contig_name], self.kmer, self.kmer_to_ids, self.canonical_k)
            node_kmers[contig_name] = kmers
            # convert kmers to numpy
            self.node_kmers = np.array(stats.zscore([node_kmers[n] for n in self.node_names], axis=0))
        self.node_lengths = np.array([node_lengths[n] for n in self.node_names])    

    def read_gfa(self):
        """Read graph file from GFA format, save list of start and end nodes to self.edges_src and self.edges_dst
        Check if contig names match the fasta file from self.node_seqs
        """
        skipped_contigs = set()
        with open(os.path.join(self.data_dir, self.graphfile), "r") as f:
            for line in f:
                if line.startswith("S"):  # sequence
                    values = line.strip().split()
                    node_name = process_node_name(values[1], self.assembly_type)
                    if node_name not in self.node_names:
                        skipped_contigs.add(node_name)
                        continue
                    contig_seq = self.node_seqs.get(node_name, "")  # discard missing contigs
                    contiglen = len(contig_seq)

                    if contiglen < self.min_contig_length:
                        skipped_contigs.add(node_name)
                    else:
                        self.graph_nodes.append(node_name)
                        # self.nodes_data.append([])
                elif line.startswith("L"):  # link/edge
                    values = line.strip().split()  # TAG, SRC, SIGN, DEST, SIGN, 0M, RC
                    src_node_name = process_node_name(values[1], self.assembly_type)
                    dst_node_name = process_node_name(values[3], self.assembly_type)
                    if src_node_name in skipped_contigs or dst_node_name in skipped_contigs:
                        # skipped_edges.add((contig_names.index(values[1]), contig_names.index(values[3])))
                        continue
                    src_index = self.node_names.index(src_node_name)
                    dst_index = self.node_names.index(dst_node_name)
                    self.edges_src.append(src_index)
                    self.edges_dst.append(dst_index)
                    if len(values) > 6:
                        rc = int(values[6].split(":")[-1])
                    else:
                        rc = 1
                    self.edge_weights.append(rc)
                    # reverse too
                    if values[1] != values[3]:
                        self.edges_src.append(dst_index)
                        self.edges_dst.append(src_index)
                        self.edge_weights.append(rc)
                elif line.startswith("P"):
                    values = line.strip().split()  # P contig path
                    self.graph_paths[values[1]] = [self.node_names.index(path_node_name[:-1]) \
                        for path_node_name in values[2].split(",") if path_node_name[:-1] in self.node_names]
        self.logger.info(f"skipped contigs {len(skipped_contigs)} < {self.min_contig_length}")
        self.adj_matrix = scipy.sparse.coo_matrix(
            (self.edge_weights, (self.edges_src, self.edges_dst)), shape=(len(self.node_names), len(self.node_names))
        )
        self.edge_weights = np.array(self.edge_weights)

    def read_gfa_contigs(self):
        skipped_contigs = set()
        edge_edge_links = {}
        contig_edge_links = {}
        with open(os.path.join(self.data_dir, self.graphfile), "r") as f:
            # ignore S entries (these are not contigs)
            # the node seqs are read from the assembly file
            for line in f:
                if line.startswith("L"): # edge links
                    values = line.strip().split()  # TAG, SRC, SIGN, DEST, SIGN, 0M, RC
                    src_node_name = process_node_name(values[1], self.assembly_type)
                    dst_node_name = process_node_name(values[3], self.assembly_type)
                    edge_edge_links.setdefault(src_node_name,set()).add(values[3])
                    edge_edge_links.setdefault(dst_node_name,set()).add(values[1])
                elif line.startswith("P"): # contig to edge links
                    values = line.strip().split()
                    contig_name = process_node_name(values[1], self.assembly_type)
                    if contig_name in skipped_contigs or contig_name not in self.node_names:
                        #skipped_edges.add((contig_names.index(values[1]), contig_names.index(values[3])))
                        continue
                    for edge in values[2].split(","):
                        contig_edge_links.setdefault(contig_name,set()).add(edge[:-1])
                    self.graph_paths[values[1]] = [self.node_names.index(path_node_name[:-1]) \
                        for path_node_name in values[2].split(",") if path_node_name[:-1] in self.node_names]
        if len(contig_edge_links) == 0:
            # load graph paths from assembly_info
            self.logger.info("reading assembly_info file for graph paths...")
            # TODO: check if file exists
            contig_edge_links = self.read_assembly_info()
        print("skipped contigs", len(skipped_contigs), "<", self.min_contig_length)
        # create graph, , and to linked to adjacent edges
        # first create edge to contig index
        edge_contig_links = {}
        for contig in contig_edge_links:
            for e in contig_edge_links[contig]:
                edge_contig_links.setdefault(e, set()).add(contig)
        for contig in contig_edge_links:
            src_index = self.node_names.index(contig)
            for edge in contig_edge_links[contig]:
                # connect contigs linked to same edge
                for contig2 in edge_contig_links[edge]:
                    dst_index = self.node_names.index(contig2)
                    if contig2 != contig:
                        self.edges_src.append(src_index)
                        self.edges_dst.append(dst_index)
                        self.edge_weights.append(1)
                        # inverse is added when we get to contig2
                for edge2 in edge_edge_links.get(edge, []):
                    for contig2 in edge_contig_links.get(edge2, []):
                        dst_index = self.node_names.index(contig2)
                        self.edges_src.append(src_index)
                        self.edges_dst.append(dst_index)
                        self.edge_weights.append(1)
        #self.logger.info(f"skipped contigs {len(skipped_contigs)} < {self.min_contig_length}")
        self.adj_matrix = scipy.sparse.coo_matrix(
            (self.edge_weights, (self.edges_src, self.edges_dst)), shape=(len(self.node_names), len(self.node_names))
        )
        self.edge_weights = np.array(self.edge_weights)

    def read_assembly_info(self, get_graph_paths=False, circular_length=100000) -> Dict[str,list]:
        """Read assembly_info.txt file generated by flye
        """
        self.circular_contigs: List[str] = []
        self.repetitive_contigs: List[str] = []
        self.contigs_multiplicity = {}
        contig_edge_links = {}
        with open(os.path.join(self.data_dir, "assembly_info.txt"), "r") as f:
            next(f) # skip header
            for line in f:
                values = line.strip().split("\t")
                contig_name = process_node_name(values[0], self.assembly_type)
                if values[3] == "Y" and int(values[1]) > circular_length:
                    self.circular_contigs.append(contig_name)
                if values[4] == "Y":
                    self.repetitive_contigs.append(contig_name)
                #if contig_name not in self.node_names:
                    #skipped_edges.add((contig_names.index(values[1]), contig_names.index(values[3])))
                #    continue
                self.contigs_multiplicity[contig_name] = int(values[5])
                # get graph_paths
                for edge in values[-1].split(","):
                    if edge != "*":
                        contig_edge_links.setdefault(contig_name,set()).add("edge_" + edge.strip("-"))
                if get_graph_paths:
                    self.graph_paths[contig_name] = ["edge_" + e for e in contig_edge_links[contig_name]]
                #for edge_name in self.graph_paths[contig_name]:
                #    self.contigs_multiplicity[self.node_names.index(edge_name)] = int(values[5])
        return contig_edge_links
    
    def print_circular_contigs(self):
        """Print circular contigs and info, based on flye
        """
        for contig_name in self.circular_contigs:
            print("found circular contig", contig_name,
                  "mult.", self.contigs_multiplicity[contig_name])
            # get edges of this contig
            contig_edges = self.graph_paths[contig_name]
            print("---", "name", "depth", "markers", "label", "edges")
            for edge_id in contig_edges:
                #edge_id = self.node_names.index("edge_" + edge)
                edge_name = self.node_names[edge_id]
                print("---", edge_id, edge_name, self.node_lengths[edge_id],
                      [round(d, 4) for d in self.node_depths[edge_id]],
                      self.node_to_label[edge_name],
                      len(self.contig_markers.get(edge_name, [])),
                      list(self.adj_matrix.getrow(edge_id).data))

    def read_depths(self) -> None:
        node_depths: Dict[str, np.array] = {}
        if self.depthfile is not None and os.path.exists(os.path.join(self.data_dir, self.depthfile)):
            if self.depthfile.endswith(".npz"):
                self.node_depths = np.load(open(os.path.join(self.data_dir, self.depthfile), "rb"))["arr_0"]
                if self.node_depths.shape[0] != len(self.node_names):
                    print("depth npz file mismatch:")
                    breakpoint()
            else:
                with open(os.path.join(self.data_dir, self.depthfile)) as f:
                    header = next(f)
                    depth_i = [i + 3 for i, n in enumerate(header.split("\t")[3:]) if "-var" not in n]
                    # var_i = [i + 3 for i, n in enumerate(header.split("\t")[3:]) if "-var" in n]
                    for line in f:
                        values = line.strip().split()
                        node_name = process_node_name(values[0], self.assembly_type)
                        if node_name in self.node_names:
                            node_depths[node_name] = np.array([float(values[i]) for i in depth_i])
                        else:
                            self.logger.info("node name not found: {}".format(node_name))
                if len(node_depths) != len(self.node_names):
                    self.logger.info("missing abundance for some nodes: {}".format(len(node_depths) - len(self.node_names)))
                self.node_depths = np.array([node_depths.get(n, [0]*len(depth_i)) for n in self.node_names])
            if len(self.node_depths[0]) > 1:  # normalize depths
                depthssum = self.node_depths.sum(axis=1) + 1e-10
                self.node_depths /= depthssum.reshape((-1, 1))
            else:
                self.node_depths = stats.zscore(self.node_depths, axis=0)
        else:
            self.node_depths = np.ones(len(self.node_names), dtype=np.float)

    def read_labels(self):
        # logging.info("loading labels from {}".format(args.labels))
        if self.labelsfile is not None:
            node_to_label = {c: "NA" for c in self.node_names}
            labels = set(["NA"])
            with open(os.path.join(self.data_dir, self.labelsfile), "r") as f:
                for line in f:
                    # label, node = line.strip().split()
                    if self.labelsfile.endswith(".csv"):
                        values = line.strip().split(",")
                    elif self.labelsfile.endswith(".tsv"):  # amber format
                        if line.startswith("@") or line.startswith("#"):
                            continue
                        values = line.strip().split("\t")
                    if len(values) < 2:
                        breakpoint()
                    node_name = process_node_name(values[0], self.assembly_type)
                    label = values[1]
                    if node_name in node_to_label:
                        node_to_label[node_name] = label
                        labels.add(label)
                    else:
                        breakpoint()
                        self.logger.info(("unused label: {}".format(line.strip())))
            labels = list(labels)
            label_to_node = {s: [] for s in labels}
            for n in node_to_label:
                s = node_to_label[n]
                label_to_node[s].append(n)
            self.node_to_label = {n: l for n, l in node_to_label.items()}
            self.labels = labels
            self.label_to_node = label_to_node
            # calculate homophily
            if len(self.edge_weights) > 1:
                self.calculate_homophily()
        else:
            self.labels = ["NA"]
            self.label_to_node = {"NA": self.node_names}
            self.node_to_label = {n: "NA" for n in self.node_names}

    def calculate_homophily(self):
        """
        Calculate percentage of edges that connect nodes with the same label
        """
        positive_edges = 0
        edges_without_label = 0
        for u, v in zip(self.edges_src, self.edges_dst):
            if self.node_to_label.get(self.node_names[u], "NA") == "NA" or \
               self.node_to_label.get(self.node_names[v], "NA") == "NA":
                edges_without_label += 1
            if self.node_to_label[self.node_names[u]] == self.node_to_label[self.node_names[v]] and self.node_to_label[self.node_names[u]] != "NA":
                positive_edges += 1
        self.logger.info(
            f"""homophily: {round(positive_edges / (len(self.edge_weights) - edges_without_label),4)} on
            {len(self.edge_weights) - edges_without_label} edges between labeled nodes"""
        )

    def read_scgs(self):
        # Load contig marker genes (Bacteria list)
        if self.scgfile is not None:
            # logging.info("loading checkm results")
            scg_path = os.path.join(self.data_dir, self.scgfile)
            ref_sets = read_marker_gene_sets(BACTERIA_MARKERS)
            contig_markers = read_contig_genes(scg_path)
            self.ref_marker_sets = ref_sets
            self.contig_markers = contig_markers
            marker_counts = get_markers_to_contigs(ref_sets, contig_markers)
            self.markers = marker_counts
        else:
            self.ref_marker_sets = {}
            self.contig_markers = {}
            self.run_checkm()

    def get_all_different_idx(self):
        """
        Returns a 2d numpy array where each row
        corresponds to a pairs of node idx whose
        feature must be different as they correspond
        to the same contig (check jargon). This
        should encourage the HQ value to be higher.
        """
        node_names_to_idx = {node_name: i for i, node_name in enumerate(self.node_names)}
        pair_idx = set()
        for n1 in self.contig_markers:
            for gene1 in self.contig_markers[n1]:
                for n2 in self.contig_markers:
                    if n1 != n2 and gene1 in self.contig_markers[n2]:
                        p1 = (node_names_to_idx[n1], node_names_to_idx[n2])
                        p2 = (node_names_to_idx[n2], node_names_to_idx[n1])
                        if (p1 not in pair_idx) and (p2 not in pair_idx):
                            pair_idx.add(p1)
        pair_idx = np.unique(np.array(list(pair_idx)), axis=0)
        print("Number of diff cluster pairs:", len(pair_idx))
        self.neg_pairs_idx = pair_idx

    def run_vamb(self, vamb_outdir, cuda, vambdim):
        from vamb.vamb_run import run as run_vamb
        self.logger.info("running VAMB")
        batchsteps = []
        vamb_epochs = 500
        if len(self.node_depths[0]) == 1:
            vamb_bs = 32
            batchsteps = [25, 75, 150]
        else:
            vamb_bs = 64
            batchsteps = [25, 75, 150, 300]
        nhiddens = [512, 512]
        self.logger.info("using these batchsteps:{}".format(batchsteps))

        vamb_logpath = os.path.join(vamb_outdir, "log.txt")
        if os.path.exists(vamb_outdir) and os.path.isdir(vamb_outdir):
            shutil.rmtree(vamb_outdir)
        os.mkdir(vamb_outdir)
        with open(vamb_logpath, "w") as vamb_logfile:
            run_vamb(
                outdir=vamb_outdir,
                fastapath=os.path.join(self.data_dir, self.fastafile),
                jgipath=os.path.join(self.data_dir, self.depthfile),
                logfile=vamb_logfile,
                cuda=cuda,
                batchsteps=batchsteps,
                batchsize=vamb_bs,
                nepochs=vamb_epochs,
                mincontiglength=self.min_contig_length,
                nhiddens=nhiddens,
                nlatent=int(vambdim),
                norefcheck=True,
            )
            if self.data_dir != "" and self.featuresfile.endswith(".tsv"):
                shutil.copyfile(os.path.join(vamb_outdir, "embs.tsv"), self.featuresfile)
            self.logger.info("Contig features saved to {}".format(self.featuresfile))

    def read_features(self):
        node_embs = {}
        self.logger.info("loading features from {}".format(self.featuresfile))
        if self.featuresfile.endswith(".tsv"):
            with open(self.featuresfile, "r") as ffile:
                for line in ffile:
                    values = line.strip().split()
                    node_embs[values[0]] = [float(x) for x in values[1:]]
            self.logger.info("loaded {} features/ {} nodes from tsv".format(len(node_embs), len(self.node_names)))
        elif self.featuresfile.endswith(".pickle"):
            with open(self.featuresfile, "rb") as ffile:
                node_embs = pickle.load(ffile)
            self.logger.info("loaded {} features/ {} nodes from pickle".format(len(node_embs), len(self.node_names)))
        self.node_embs = [
                node_embs.get(n, np.random.uniform(10e-5, 1.0, len(node_embs[list(node_embs.keys())[0]]))) for n in self.node_names
            ]  # deal with missing embs
        self.node_embs = np.array(self.node_embs)
        

    def get_topk_neighbors(self, k, scg_only=False):
        """
        Returns a list of the top k neighbors for each node. Use kmers and abundance

        """
        #breakpoint()
        self.logger.info("getting top {} neighbors".format(k))
        self.topk_neighbors = []
        features = np.concatenate((self.node_kmers, self.node_depths), axis=1)
        cosine_dists = np.dot(features, features.T)
        for i in range(len(self.node_names)):
            self.topk_neighbors.append(set(np.argsort(cosine_dists[i])[-k:]))
        self.logger.info("got top {} neighbors".format(k))

    def estimate_n_genomes(self):
        self.scg_counts = {}
        for marker_set in self.ref_marker_sets:
            for gene in marker_set:
                self.scg_counts[gene] = 0
                for contig in self.contig_markers:
                    if gene in self.contig_markers[contig]:
                        self.scg_counts[gene] += self.contig_markers[contig][gene]

        #print(self.scg_counts)
        quartiles = np.percentile(list(self.scg_counts.values()), [25, 50, 75])
        print("candidate k0s", sorted(set([k for k in self.scg_counts.values() if k >= quartiles[2]])))
        return max(self.scg_counts.values())
        
       
    def run_checkm(self, nthreads=10, tempdir="../temp"):
        # check if checkm is installed
        checkm_is_avail = shutil.which("checkm") is not None
        # if not, print commands only
        commands = [
            "mkdir {}/nodes".format(self.data_dir), #; cd nodes; ",
            "cat {0}/{1} | awk '{{ if (substr($0, 1, 1)=='>') {{filename=(substr($0,2) '.fa')}} print $0 > {0}/filename }}".format(self.data_dir, self.fastafile),
            'find {}/nodes/ -name "* *" -type f | rename "s/ /_/g"'.format(self.data_dir),
            "checkm taxonomy_wf --tmpdir {0} -t {1} -x fa domain Bacteria {2}/nodes/ {2}/checkm_nodes/".format(tempdir, nthreads, self.data_dir)
        ]
        if checkm_is_avail:
            #run commands one by one
            for cmd in commands:
                os.system(cmd)
        else:
            print("Run this on a machine with CheckM installed")
            for cmd in commands:
                print(cmd)


    def read_gtdbtk_files(self, prefix="gtdbtk", suffix_labels="summary.tsv",
                          suffix_markers="markers_summary.tsv", reset=True):
        # use summary to add species level labels
        # also save full lineages
        roots = ["bac120"] #, "ar53"]
        no_species = 0
        no_genus = 0
        species_labels = set()
        # reset labels
        if reset:
            self.node_to_label = {n: "NA" for n in self.node_names}
            self.labels = ["NA"]
        contig_markers = {n: {} for n in self.node_names}
        ref_markers = set()
        for root in roots:
            # read labels
            with open(os.path.join(self.data_dir, f"{prefix}.{root}.{suffix_labels}"), 'r') as f:
                next(f) # skip header
                for line in f:
                    values = line.strip().split("\t")
                    node_name = process_node_name(values[0], self.assembly_type)
                    if values[1].startswith("Unclassified"):
                        continue # do not add unclassified nodes
                    classification = values[1].split(";")
                    node_slabel = classification[-1]
                    node_glabel = classification[-2]
                    #if values[5] != "N/A":
                    #    fast_ani = float(values[5])
                    #    anis.append(fast_ani)
                    if node_slabel == "s__":
                        no_species += 1
                    else:
                        self.node_to_label[node_name] = node_slabel
                        species_labels.add(node_slabel)
                    #if node_slabel == "g__":
                    #    no_genus += 1
                    #else:
                    #    node_to_glabel[node_name] = node_glabel
                    #    genus_labels.add(node_glabel)

            # read markers
            with open(os.path.join(self.data_dir, f"{prefix}.{root}.{suffix_markers}"), 'r') as f:
                next(f) # skip header
                for line in f:
                    values = line.strip().split("\t")
                    node_name = process_node_name(values[0], self.assembly_type)
                    if int(values[1]) > 0:
                        contig_markers[node_name] = {g: 1 for g in values[5].split(",")}
                        ref_markers.update(set(contig_markers[node_name]))
                    if int(values[4]) > 0:
                        ref_markers.update(set(values[-1].split(",")))
        # create ref labels and ref markers
        self.labels += list(species_labels)
        self.label_to_node = {s: [] for s in self.labels}
        for n in self.node_to_label:
            s = self.node_to_label[n]
            self.label_to_node[s].append(n)
        assert len(ref_markers) == 120 # GTDB-tk bac and ar markers
        if reset:
            self.ref_marker_sets = [ref_markers]    
            self.contig_markers = contig_markers
            marker_counts = get_markers_to_contigs(self.ref_marker_sets, contig_markers)
            self.markers = marker_counts

    def get_edges_with_same_scgs(self):
        """Check every edge to see if nodes have SCGs in common

        :return: edges IDs (edges_src, edges_dst, edge_weight)
        :rtype: list
        """
        edge_ids_with_same_scgs = []
        scg_counter = Counter()
        for x, (i, j) in enumerate(zip(self.edges_src, self.edges_dst)):
            if self.node_names[i] in self.contig_markers and self.node_names[j] in self.contig_markers and \
                len(self.contig_markers[self.node_names[i]]) > 0 and len(self.contig_markers[self.node_names[j]]) > 0:
            
                overlap = len(self.contig_markers[self.node_names[i]].keys() & \
                    self.contig_markers[self.node_names[j]].keys())
                if overlap > 0 and i != j:
                    #remove edge
                    scg_counter[overlap] += 1
                    edge_ids_with_same_scgs.append(x)
        print("edges with overlapping scgs (max=20):", scg_counter.most_common(20))
        return edge_ids_with_same_scgs

    def generate_edges_based_on_labels(self, noise=0):
        # link nodes with same labels and create new adj_matrix/edges_src/edges_dst/edge_weights
        #new_edges = []
        new_srcs = []
        new_dsts = []
        for label in self.label_to_node:
            # make circular graph
            for i in range(len(self.label_to_node[label])):
                new_src = self.node_names.index(self.label_to_node[label][i])
                new_srcs.append(new_src)
                if i+1 < len(self.label_to_node[label]):
                    new_dst = self.node_names.index(self.label_to_node[label][i+1])
                else: # connect to element 0
                    new_dst = self.node_names.index(self.label_to_node[label][0])
                #new_edges.append((new_src, new_dst))
                new_dsts.append(new_dst)
        new_weights = [1] * len(new_srcs)
        print("creating new graph with {} edges".format(len(new_srcs)))
        self.adj_matrix = scipy.sparse.coo_matrix(
            (new_weights, (new_srcs, new_dsts)), shape=(len(self.node_names), len(self.node_names))
        )
        self.edges_src = new_srcs
        self.edges_dst = new_dsts
        self.edge_weights = np.array(new_weights)



